; TITLE: 8051 eForth ************************************************
; rewritten into UCASM format 30.07.92 by Wolfgang Schemmert
; Version upcount into 1.02
; words changed in comparison with version V1.01
;       ?STACK
;       TOKEN
;       $,n
;       COLD

; NEW WORDS in Version V1.02:

; CONSTANT ( w --  ; <string>)
;       Compile a new constant initialized with value w

; DO
;       Alias of FOR. Begins a count down loop.

; FLIP     ( -- )
;       Swap both bytes of the TOS

; FORGET   ( --; <string> )
;       Erase the word with name <string> and all following from dictionary
;       Reset current dictionary pointer to last remaining word.
;       More sophisticated dictionary handling not yet implemented.

;FUNCTION ( a -- ; <string> )
;       Compile a call for external function.
;       The executable machine code must begin at address a.
;       Back to eForth at the first "ret" instruction found.
;       Free use of 8031 Accu, Register Banks 2 and 3,
;       hardware stack, if left with aligned stackpointer.
;       Free use of SFR's, memory, subroutines and interrupts
;       as long as not destroing eForth system & setup.
;       Internal RAM lower than $10 is forbidden, because eForth
;       pointers are stored there.
;       Hardware stack is in internal RAM, from $40 growing upwards.
;       This area is forbidden, too.
;       When calling "FUNCTION", automatically register bank #2
;       is activated; after "ret" selection is done back to bank #1.

; JOIN   ( wHI wLO -- w )
;       Joins the lower bytes of the two stack top items
;       to one word held in TOS

; LOADBIN  ( a u -- )
;       Loads u arbitrary bytes to address a via SerialCom.
;       No semantic or allocation precautions, use only if you
;       are looking through.

; LOADSYSTEM  ( -- )
;       Load a precompiled dictionary extension from the terminal.
;       USER variables are refreshed, too.
;       Data format see "SAVESYS"
;       During loading, RAMBOT and UPP are compared with present
;       values. If inadequate, loading is refused. ( Echo ESC)
;       If OK, further loading is triggered by echoing PACE
;       LAST is inherently restored by loading the user area.
;       VARAD is temporarily saved before loading. If the previous
;       value of VARAD is bigger than the loaded one, system VARAD is
;       restored to the user area, assuming other users have defined
;       additional public variables. The values of public variables
;       are not changed by LOADSYS.

; MS    ( u -- )
;       Delay program execution for about u milliseconds.
;       If u == 0, then delay is about 40 microseconds.
;       Note, that delay time is only correct for 11.059MHz quartz !

; SAVEBIN  ( a u -- )
;        pushes u arbitrary bytes starting from address a
;        via Serial Com. Uses XON/XOFF protocol. Simple data
;        format without header nor security.

; SAVESYSTEM  ( -- )
;        save the dictionary extension beeing actually in the
;        RAM area via SerialCom. Present protocol: binary byte
;        stream without security precautions but using XON/XOFF protocol.
;        Length counting is provided inherently. Total length of file
;        is:   (CP-RAMBOT) + USOFF + 8  bytes
;        Data format: (WORDs are transmitted INTEL-style LOfirst-HInext)
;               .WORD   RAMBOT
;               .WORD   UPP
;               .WORD   CP (last byte used by dictionary extension)
;               .WORD   USOFF
;               .BYTEs   dictionary extension from RAMBOT to CP
;               .BYTEs   user area from UPP to (UPP+USOFF)
;        Attention: information about LAST and VARAD is contained
;        in the user block!

; P1!   ( c -- )
;       Pop the data stack low byte to SFR "P1" (I/O port)

; P1@    ( -- c )
;       Push contents of SFR "P1" to the data stack.

; P3!   ( c -- )
;       Pop the data stack low byte to SFR "P3" (I/O port).

; P3@    ( -- c )
;       Push contents of SFR "P3" to the data stack.

; SFR!  ( c b8 -- )
;       Pop the data stack to processor direct memory or SFR.
;       (One byte popped to 8Bit address,
;       both held in LowByte of stack words)

; SFR@    ( b8 -- c )
;       Push SFR or direct RAM (byte memory location) to the data stack.
;       (One byte pushed from 8Bit address,
;       both held in LowByte of stack words)

; SPLIT  ( w -- wHI wLO)
;        splits the word in TOS into two words,
;        each containing one byte with leading zeroes

; USA      ( -- ;  <string> )
;        create a USer variable with Automatic offset increment
;        (Berlinian slang of "USER" in german pronunciation)

; USOFF    ( -- a)
;        user variable holds updated offset for USA

; VARAD    ( -- a)
;        user variable holds PFA of next
;        defineable variable in extra RAM portion.
;        necessary to make dictionary extensions
;        burnable into EEPROM

; ZLOOP
;        Alias of "NEXT", terminates a count-down loop structure.
;        It is different from the Forth-83 "LOOP" as far as it
;        always counts down the start parameter until it
;        has reached zero.

;********************************************************************

;Rev 14.10.92: LISTT (DOLST in reverse byte order) removed from source
;              and DOLIT,LISTT replaced by DOLIT,DOLST,FLIP
;              new words added: MS,P1!,P1@,P3!,P3@,SFR!,SFR@,FUNCTION
;              "COLDINIT" and reset values for CP and NP changed during
;              implementation of SFR! and SFR@.

;Rev 11.10.92: new word: FORGET

;Rev 09.10.92: BASE: HEX =16
;              USOFF and VARAD replaced within the USER-Field
;              FLIP,SLPIT,JOIN,SAVESYSTEM,LOADSYSTEM,SAVEBIN,LOADBIN new

;Rev 26.09.92: LORAM renamed into RAMBOT  for better Mnemonics
;              USOFF and USA coded for automatic allocation
;                                        of USER-Variables
;              VARAD new
;              prinzipal chgange of storage mechanism in VARIABLE
;              now into RAM-segment below data stack
;              So, even dictionary extensions become ROM able.
;              CONSTANT introduced

;Rev 16.09.92: RAM-area relocated $4000 to $5FFF

;Rev 31.08.92: Constant LORAM is beginning address of RAM
;              CP and NP initialised appropriately
;              Base = DECIMAL = 10

;Rev 23.08.92:"FILE" corrected: parts were gone lost during work

;================================================================
;   Editorial note of the original Distribution of eForth V1.01 :
;================================================================
;
;	8051 eForth 1.1 by C. H. Ting, 1990
;
;	This eForth system was developed using chipForth from Forth, Inc.
;	The eForth Model was developed by Bill Muench and C. H. Ting.
;
;	The goal of this implementation is to show that the eForth Model
;	can be ported to a ROM based 8 bit microprocessor, Intel 8051.
;	Deviations from the original eForth Model are:
;
;       All kernel words are assembled as DB statements.
;       Memory map is tailored to a ROM based system.
;       $COLON and $USER are modified to compile LJMP doLIST.
;       call, compiles a LCALL with a flipped destination address.
;       USER, VARIABLE and : are modified to use above 'call,'.
;       FORTH vocabulary pointer is a pair user variables.
;       BYE is deleted.
;
;   To assemble this source file and generate a ROM image,
;   type the following commands using MASM and LINK:
;       >MASM 8051;
;       >LINK 8051;
;   The resulting 8051.EXE contains the binary image suitable
;   for PROM programming.  The actual image is offset by 200H
;   bytes from the beginning of the .EXE file.  This image
;   must be placed in a PROM from 0 to 1FFFH, and it uses a RAM
;   chip from 8000H to 9FFFH.  If your system does not have
;   this memory configuration, modify the memory pointers in
;   the source file accordingly.  Places to be modified are
;   marked by '******'.
;
;	8051 is a slow processor.  Do not expect great performance
;	of this implementation, considering that most words are in high
;	level.  Your are encouraged to recode some of the high level words
;	to optimize its performance.
;
;	Direct your questions and contributions to:
;
;		Dr. C. H. Ting
;		156 14th Avenue
;		San Mateo, CA 94402
;		(415) 571-7639
;
;================================================================
; In this release an alternative assembly method is used :
;
;       To assemble this source file use "UCASM" Assembler !
;       The assembler directive .DW means storage of a 16Bit word in the
;       Intel-style order LOByte - HIByte.
;       Because UCASM does know neither assembly variables nor forward
;       references of memory allocation, some changes in the setup
;       strategy became neccesary
;================================================================

; Version control

VER     .EQU     $01                     ;major release version
EXT     .EQU     $02                     ;minor extension

; machine-specific addresses and Special-Function-Registers

REG0    .EQU  0     ;Hilfsregister
REG1    .EQU  1
RPH     .EQU  4     ;Forth-Return-Pointer
SPH     .EQU  5     ;Daten-Stack-Pointer
UPL     .EQU  6     ;User-Pointer
UPH     .EQU  7

SPL     .EQU  8     ;
RPL     .EQU  9     ;
TPL     .EQU  10    ;Top of Data Stack
TPH     .EQU  11    ;

;(W.S.) order of registers found actually implemented different from
;paper describing the eForth model

NPL     .EQU  12    ;Notepad (Scratch Register)
NPH     .EQU  13
IPL     .EQU  14    ;Instruction Pointer
IPH     .EQU  15

SP      .EQU  $81
DPL     .EQU  $82
DPH     .EQU  $83
PCON    .EQU  $87
TCON.6  .BIT  $8E   ;Start/Stop Timer1
TMOD    .EQU  $89
TH1     .EQU  $8D
PORT1   .EQU  $90
PORT3   .EQU  $B0
SCON    .EQU  $98
SCON.0  .BIT  $98   ;Empfangs-Interrupt
SCON.1  .BIT  $99   ;Sende-Interrupt
SBUF    .EQU  $99
P2      .EQU  $A0
IE      .EQU  $A8
IE.4    .BIT  $AC   ;Freigabe Interrupt Ser.Port
PSW     .EQU  $D0   ;Programm-Status-Wort
ACC.7   .BIT  $E7



; Constants
COMPO   .EQU  $40       ;lexicon compile only bit
IMEDD   .EQU  $80       ;lexicon immediate bit
MASKK   .EQU  $7F1F     ;lexicon bit mask
			;mask out LEX and bit7 of first name character
			;during dictionary search

CELLL   .EQU  2         ;size of a cell !!! Must not be changed !!!
BASEE   .EQU  16        ;HEX is default radix
; VOCSS (depth of vocabulary stack): must be coded directly into user scheme

BKSPP   .EQU  8         ;backspace
LF      .EQU  10        ;line feed
CRR     .EQU  13        ;carriage return
ERR     .EQU  27        ;error escape
TIC     .EQU  39        ;tick

; Memory allocation    0//name&code>--//--<ROM|RAM>--<sp//tib>--rp//em

RAMTOP  .EQU  $6000     ;top of RAM memory******
RAMBOT  .EQU  $4000     ;bottom of RAM memory*******
BM      .EQU  0         ;bottom of ROM memory******
INITCODE .EQU  BM+$40   ;cold start vector******

US      .EQU  $0100     ;user area size in cells
RTS     .EQU  $0100     ;return stack/TIB size
DTS     .EQU  $0100     ;data stack size

UPP     .EQU  RAMTOP-US ;start of user area (UP0)
TIBB    .EQU  UPP-RTS   ;terminal input buffer (TIB)
RPP     .EQU  UPP-2     ;start of return stack (RP0)
SPP     .EQU  RPP-RTS   ;start of data stack (SP0)
SFRCODE .EQU  SPP-DTS-14;special code segment to be dynamically
			;reassembled for access to SFR's, which can
			;only be handled in the direct addressing mode
VARAM   .EQU  SPP-DTS-16;start of VARIABLE PFA RAM (growing downwards)

; (WS) I omitted the originally implemented concept, which uses an extra
; segment of headers. In my eyes, that brings some complications but no
; additional comfort or features. Now, like in conventional Forth systems,
; headers und body are threaded interlacing in a unique dictionary segment.

; Header construction:
;               Align HEADER to even address
;               .DW  ca
;       (la:)   .DW  (LILABEL = na) of previous word in this vocabulary
;     LILABEL:
;       (na:)   .DB  lexicon byte    (flags | count)
;               .DB  'namestring'
;               Align CODE to even address
;       LABEL:
;       (ca:)   begin of code
;               .
;               .
;               .
;               end of code
;               next header begins with "Align HEADER to even address...."

; .DW means 16 Bit word stored in the Intel-style-order LOWbyte-->HIGHbyte


; Define assembly macros

; Adjust an address to the next cell boundary.
	.MACRO @ALIGN
		.ALIGN 2
	.ENDM

; Compile a code definition header.
	.MACRO @CODE PLABEL,PLILABEL,LINK,LEX,NAME
		@ALIGN                  ;force HEADER to even address
		.DW ( * + 6+((LEX) & $1E)) ;points to subsequent code-address
		.DW LINK                ;Link to previous name-address
PLILABEL:       .DB  LEX,NAME           ;name string
		@ALIGN                  ;force CODE to even address
PLABEL:
					;Begin of CODE
	 .ENDM

; Compile a colon definition header.
; (WS)see Macro definition below "doLIST"

; Compile a user variable header.
; (WS)see Macro definition below "doUSER"

; Compile an inline string.
; (WS)LEN had to be added to the original version, because UCASM does not
; support assembly variables

	.MACRO   @DISTR FUNCT,LEN,STRNG
		.DW     FUNCT           ;function
		.DB      LEN,STRNG      ;count byte and string
		@ALIGN
	.ENDM


; Main entry points and COLD start data

	.CODE BM        ;Power up location******

	ljmp ORIG  ;Jump to cold start

	.DB      $32,0,0,0,0,0,0,0               ;Return from interrupt
	.DB      $32,0,0,0,0,0,0,0               ;Return from interrupt
	.DB      $32,0,0,0,0,0,0,0               ;Return from interrupt
	.DB      $32,0,0,0,0,0,0,0               ;Return from interrupt
	.DB      $32,0,0,0,0,0,0,0               ;Return from interrupt
	.DB      $32,0,0,0,0,0,0,0               ;Return from interrupt
	.DB      $32,0,0,0,0,0,0,0               ;Return from interrupt
	.DB      $32,0,0,0,0


; (WS) at this point, automatically  * = COLDD
	.ENDS

;(WS)Da _USER nicht als Assemblierungsvariable inkrementiert werden kann
;explizite Benennung der User-Variablen wie ihre Dictionary-Labels
;mit vorangestelltem U
;UCASM hat Problem mit Vorwrtsreferenzen bei .EQU Zuweisungen.
;daher wird abweichend vom Original die Zuweisung der USER-Variablen
;als Assemblercode bei COLDENTRY durchgefhrt.


	.DATA UPP

; (WS) intermezzo: define RAM labels for USER variables

USZERO: .DW             ;SP0
URZERO: .DW             ;RP0
UTQKEY: .DW             ;'?KEY
UTEMIT: .DW             ;'EMIT
UTEXPE: .DW             ;'EXPECT
UTTAP:  .DW             ;'TAP
UTECHO: .DW             ;'ECHO
UTPROM: .DW             ;'PROMPT
UBASE:  .DW             ;BASE
UTEMP:  .DW             ;tmp
USPAN:  .DW             ;SPAN
UINN:   .DW             ;>IN
UNTIB:  .DW             ;#TIB
UTIP:   .DW             ;TIBU
UCSP:   .DW             ;CSP
UTEVAL: .DW             ;'EVUL
UTNUMB: .DW             ;'NUMBER
UHLD:   .DW             ;HL
UHANDL: .DW             ;HANDLER
; next two new in this version:
UVARAD: .DW             ;next approproate variable PFA in RAM
UUSOFF: .DW             ;next appropriate user offset

UCNTXT: .DW             ;CONTEXT pointer
UVOCS:  .BLOCK 16       ;vocabulary stack for max. 8 vocabularies
UCRRNT: .DW             ;CURRENT pointer
UVOCL:  .DW             ;vocabulary link pointer
UCP:    .DW             ;RAMBOT ***** ;now: dictionary pointer into RAM
			;(headers & code interleaved /growing addr.)
UNP:    .DW             ;pointer to next name field. Reset CP to this value
			;in "THROW". Other meaning than in original eForth
ULAST:  .DW             ;LAST
UVFRTH: .DW             ;FORTH
UVL:    .DW             ;vocabulary link

	.ENDS

	.CODE INITCODE  ;go on with executable code

ORIG:	;Cold boot routine
	;Part: hardware and pointer initialisation

; Timer 1 auf kont.Teilung /3 schalten und SMOD (PCON Bit7) setzen
; entspr. 9600 Baud b.Quarz 11.059MHz
	mov TH1,#$fd        ;Timer laden mit Teilerverh. /3
	mov PCON,#0         ;0(-->9600 Baud) $80(--> 19200 Baud)
	mov SCON,#$50       ;UART Empfang einschalten, Modus 8N1
	setb TCON.6         ;Timer starten
	mov TMOD,#$20

	mov IE,#0
	mov SP,#$40     ;processor hardware return stack
	mov PSW,#8      ;Register Bank 1 aktivieren
	mov DPTR,#RPP
	mov RPL,DPL
	mov RPH,DPH
	mov DPTR,#SPP
	mov SPL,DPL
	mov SPH,DPH

	mov P2,SPH          ;Adre-Segment auf Stackpointer


	lcall GETCOLD1
;       this is necessary for the assembler to avoid a forward reference
;       set instruction pointer to high-level "COLD"
	mov R6,DPL
	mov R7,DPH

;  ... and now go on with RETURN. At this point there is a change
;      from running "normal" machine code to the processing
;      of threaded code. The virtual FORTH machine is lifting off.


RETURN: ;The Forth Inner Interpreter

	mov DPL,R6      ;IPL                    2 Cycles
	mov DPH,R7      ;IPH                    2 Cycles
	clr A                                 ; 1 Cycle
	movc A,@A+DPTR                        ; 2 Cycles
	mov NPL,A       ;NPL                    1 Cycle
	mov A,#1                              ; 1 Cycle
	movc A,@A+DPTR                        ; 2 Cycles
	mov DPH,A                             ; 1 Cycle
	mov DPL,R4      ;NPL                    2 Cycles
	mov A,R6        ;IPL                    1 Cycle
	add A,#2                              ; 1 Cycle
	mov IPL,A       ;IPL                    1 Cycle
	jnc RETURN1                           ; 2 Cycles
	inc R7          ;IPH                    1 Cycle
RETURN1:
	clr A                                 ; 1 Cycle
	jmp @A+DPTR                           ; 2 Cycles
					;total 23 Cycles

; The kernel


; doLIT ( -- w )
;       Push an inline literal.

	@CODE DOLIT,LDOLIT,0,COMPO+5,'doLIT'
	mov DPL,R6      ;R6
	mov DPH,R7      ;R7
	mov A,R3        ;TPH
	movx @R0,A      ;SPL
	dec R0          ;SPL
	mov A,R2        ;TPL
	movx @R0,A      ;SPL
	dec R0          ;SPL
	clr A
	movc A,@A+DPTR
	mov R2,A        ;TPL
	inc DPTR
	clr A
	movc A,@A+DPTR
	mov R3,A        ;TPH
	inc DPTR
	mov R6,DPL      ;IPL
	mov R7,DPH      ;IPH
	ajmp RETURN+4


; test  (--)
;       moves TOS directly to serial TX; special system startup debugging

	@CODE TEST,LTEST,LDOLIT,4,'TEST'
	mov A,TPH
	lcall senda
	mov A,TPL
	lcall senda
	mov a,#$20
	lcall senda
	ajmp RETURN


; doLIST ( a -- )
;       Process colon list.

	@CODE DOLST,LDOLST,LTEST,COMPO+6,'doLIST'
	mov P2,RPH      ;Get list address
	mov A,R7
	movx @R1,A      ;RPL
	dec R1          ;RPL
	mov A,R6        ;IPL
	movx @R1,A      ;RPL
	dec R1          ;RPL
	mov P2,SPH      ;Restore stack pointer
	pop IPH
	pop IPL
	ajmp RETURN


	.MACRO  @COLON  SLABEL,SLILABEL,LINK,LEX,NAME
		@ALIGN                  ;force HEADER to even address
		.DW ( * + 6+((LEX) & $1E)) ;points to subsequent code-address
		.DW LINK                ;Link to previous name-address
SLILABEL:       .DB  LEX,NAME           ;name string
		@ALIGN                  ;force CODE to even address
SLABEL:
					;Begin of CODE
		nop                     ;align dummy
		lcall DOLST             ;Forth ENTER
	.ENDM


; next  ( -- )
;       Run time code for the single index loop.
;       : next ( -- ) \ hilevel model
;       r> r> dup if 1 - >r @ >r exit then drop cell+ >r ;

	@CODE DONXT,LDONXT,LDOLST,COMPO+4,'NEXT'
	mov P2,RPH
	inc R1          ;RPL
	movx A,@R1      ;RPL
	clr C
	subb A,#1
	movx @R1,A      ;RPL
	inc R1          ;RPL
	movx A,@R1      ;RPL
	subb A,#0
	movx @R1,A      ;RPL
	mov P2,SPH
	mov DPL,R6      ;IPL
	mov DPH,R7      ;IPH
	jnc NEXT1
	inc DPTR
	inc DPTR
	mov R6,DPL      ;IPL
	mov R7,DPH      ;IPH
	ajmp RETURN
NEXT1: dec R1           ;RPL
	dec R1          ;RPL
	clr A
	movc A,@A+DPTR
	mov R6,A        ;IPL
	mov A,#1
	movc A,@A+DPTR
	mov R7,A        ;IPH
	ajmp RETURN


; ?branch ( f -- )
;       Branch if flag is zero.

	@CODE QBRAN,LQBRAN,LDONXT,COMPO+7,'?branch'
	mov DPL,R6      ;IPL
	mov DPH,R7      ;IPH
	mov A,R2        ;TPL
	orl A,R3        ;TPH
	jz QBRAN1
	inc DPTR
	inc DPTR
	mov R6,DPL      ;IPL
	mov R7,DPH      ;IPH
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R2,A        ;TPL
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R3,A        ;TPH
	ajmp RETURN
QBRAN1:  clr A
	movc A,@A+DPTR
	mov R6,A        ;IPL
	mov A,#1
	movc A,@A+DPTR
	mov R7,A        ;IPH
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R2,A        ;TPL
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R3,A        ;TPH
	ajmp RETURN


; BRANCH  ( -- )
;       Branch to an inline address.

	@CODE   BRAN,LBRAN,LQBRAN,COMPO+6,'BRANCH'
	mov DPL,R6      ;IPL
	mov DPH,R7      ;IPH
	clr A
	movc A,@A+DPTR
	mov R6,A        ;IPL
	mov A,#1
	movc A,@A+DPTR
	mov R7,A        ;IPH
	ajmp RETURN


; EXECUTE ( ca -- )
;       Execute the word at ca.

	@CODE EXECU,LEXECU,LBRAN,7,'EXECUTE'
	mov DPL,R2      ;TPL
	mov DPH,R3      ;TPH
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R2,A        ;TPL
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R3,A        ;TPH
	clr A
	jmp @A+DPTR


; EXIT  ( -- )
;       Terminate a colon definition.

	@CODE EXIT,LEXIT,LEXECU,4,'EXIT'
	mov P2,RPH
	inc R1          ;RPL
	movx A,@R1      ;RPL
	mov R6,A        ;IPL
	inc R1          ;RPL
	movx A,@R1      ;RPL
	mov R7,A        ;IPH
	mov P2,SPH
	ajmp RETURN


; MS    ( u -- )
;       Delay program execution for about u milliseconds.
;       If u == 0, then delay is about 40 microseconds.
;       Note, that delay time is only correct for 11.059MHz quartz !

	@CODE MS,LMS,LEXIT,2,'MS'
	mov A,R2        ;TPL                      1 Cycle
	orl A,R3        ;TPH                      1 Cycle
	jz MS3          ;forget, if TIME = 0      2 Cycles
	inc TPH         ;to compensate countdown before compare
MS1:    mov A,#76
MS2:    nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	dec A
	jnz MS2         ;12 Cycles total is 13Micros
	nop
	nop
	nop
	nop
	mov A,TPL
	dec A
	mov TPL,A
	jnz MS1
	mov A,TPH
	dec A
	mov TPH,A
	jnz MS1
MS3:
	inc R0          ;SPL     ;restore TOS     1 Cycle
	movx A,@R0      ;SPL                      2 Cycles
	mov R2,A        ;TPL                      1 Cycle
	inc R0          ;SPL                      1 Cycle
	movx A,@R0      ;SPL                      2 Cycles
	mov R3,A        ;TPH                      1 Cycle
	ajmp RETURN                             ; 2 Cycles

; !     ( w a -- )
;       Pop the data stack to memory.

	@CODE STORE,LSTORE,LMS,1,'!'
	mov DPL,R2      ;TPL
	mov DPH,R3      ;TPH
	inc R0          ;SPL
	movx A,@R0      ;SPL
	movx @DPTR,A
	inc DPTR
	inc R0          ;SPL
	movx A,@R0      ;SPL
	movx @DPTR,A
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R2,A        ;TPL
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R3,A        ;TPH
	ajmp RETURN


; @     ( a -- w )
;       Push memory location to the data stack.

	@CODE AT,LAT,LSTORE,1,'@'
	mov DPL,R2      ;TPL
	mov DPH,R3      ;TPH
	movx A,@DPTR
	mov R2,A        ;TPL
	inc DPTR
	movx A,@DPTR
	mov R3,A        ;TPH
	ajmp RETURN


; C!    ( c b -- )
;       Pop the data stack to byte memory.

	@CODE CSTOR,LCSTOR,LAT,2,'C!'
	mov DPL,R2      ;TPL
	mov DPH,R3      ;TPH
	inc R0          ;SPL
	movx A,@R0      ;SPL
	movx @DPTR,A
	inc R0          ;SPL
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R2,A        ;TPL
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R3,A        ;TPH
	ajmp RETURN


; C@    ( b -- c )
;       Push byte memory location to the data stack.

	@CODE CAT,LCAT,LCSTOR,2,'C@'
	mov DPL,R2      ;TPL
	mov DPH,R3      ;TPH
	movx A,@DPTR
	mov R2,A        ;TPL
	mov R3,#0       ;TPH
	ajmp RETURN


; SFR!  ( c b8 -- )
;       Pop the data stack to processor direct memory or SFR.
;       (One byte to 8Bit address, both held in LowByte of stack words)

	@CODE SFRST,LSFRST,LCAT,4,'SFR!'
	mov DPTR,#(SFRCODE+1)
	mov A,TPL       ;dynamically compile SFR address
	movx @DPTR,A
	inc R0          ;SPL
	movx A,@R0      ;SPL    data to be written
	lcall SFRCODE
	inc R0          ;SPL    drop HiByte
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R2,A        ;TPL
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R3,A        ;TPH
	ajmp RETURN


; SFR@    ( b8 -- c )
;       Push SFR or direct RAM (byte memory location) to the data stack.
;       (One byte to 8Bit address, both held in LowByte of stack words)

	@CODE SFRAT,LSFRAT,LSFRST,4,'SFR@'
	mov DPTR,#(SFRCODE+4)
	mov A,TPL       ;dynamically compile SFR address
	movx @DPTR,A
	lcall (SFRCODE+3)
	mov TPL,A       ;get contents of SFR or direct RAM
	mov R3,#0       ;clear TPH
	ajmp RETURN


; P1!   ( c -- )
;       Pop the data stack low byte to SFR "P1".

	@CODE P1STO,LP1STO,LSFRAT,3,'P1!'
	mov PORT1,TPL
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R2,A        ;TPL
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R3,A        ;TPH
	ajmp RETURN


; P1@    ( -- c )
;       Push contents of SFR "P1" to the data stack.

	@CODE P1AT,LP1AT,LP1STO,3,'P1@'
	mov A,R3        ;TPH
	movx @R0,A      ;SPL
	dec R0          ;SPL
	mov A,R2        ;TPL
	movx @R0,A
	dec R0          ;SPL
	;mov PORT1,#$FF  ;first must be polarized
	mov TPL,PORT1
	mov R3,#0       ;TPH
	ajmp RETURN


; P3!   ( c -- )
;       Pop the data stack low byte to SFR "P3".

	@CODE P3STO,LP3STO,LP1AT,3,'P3!'
	mov PORT3,TPL
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R2,A        ;TPL
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R3,A        ;TPH
	ajmp RETURN


; P3@    ( -- c )
;       Push contents of SFR "P3" to the data stack.

	@CODE P3AT,LP3AT,LP3STO,3,'P3@'
	mov A,R3        ;TPH
	movx @R0,A      ;SPL
	dec R0          ;SPL
	mov A,R2        ;TPL
	movx @R0,A
	dec R0          ;SPL
	;mov PORT3,#$FF  ;first must be polarized
	mov TPL,PORT3
	mov R3,#0       ;TPH
	ajmp RETURN


; >R    ( w -- )
;       Push the data stack to the return stack.

	@CODE TOR,LTOR,LP3AT,COMPO+2,'>R'
	mov P2,RPH
	mov A,R3        ;TPH
	movx @R1,A      ;RPL
	dec R1          ;RPL
	mov A,R2        ;TPL
	movx @R1,A      ;RPL
	dec R1          ;RPL
	mov P2,SPH
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R2,A        ;TPL
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R3,A        ;TPH
	ajmp RETURN


; R@    ( -- w )
;       Copy top of return stack to the data stack.

	@CODE RAT,LRAT,LTOR,2,'R@'
	mov A,R3        ;TPH
	movx @R0,A      ;SPL
	dec R0          ;SPL
	mov A,R2        ;TPL
	movx @R0,A
	dec R0          ;SPL
	mov DPL,R1      ;RPL
	mov DPH,RPH
	inc DPL
	movx A,@DPTR
	mov R2,A        ;TPL
	inc DPL
	movx A,@DPTR
	mov R3,A        ;TPH
	ajmp RETURN


; R>    ( -- w )
;       Pop the return stack to the data stack.

	@CODE RFROM,LRFROM,LRAT,2,'R>'
	mov A,R3        ;TPH
	movx @R0,A
	dec R0          ;SPL
	mov A,R2        ;TPL
	movx @R0,A
	dec R0          ;SPL
	mov P2,RPH
	inc R1          ;RPL
	movx A,@R1      ;RPL
	mov R2,A        ;TPL
	inc R1          ;RPL
	movx A,@R1
	mov R3,A        ;RPH
	mov P2,SPH
	ajmp RETURN


; RP@   ( -- a )
;       Push the current RP to the data stack.

	@CODE RPAT,LRPAT,LRFROM,3,'RP@'
	mov A,R3        ;TPH
	movx @R0,A      ;SPL
	dec R0          ;SPL
	mov A,R2        ;TPL
	movx @R0,A
	dec R0          ;SPL
	mov TPL,R1      ;TPL,RPL
	mov R3,RPH      ;TPH
	ajmp RETURN


; RP!   ( a -- )
;       Set the return stack pointer.

	@CODE RPSTO,LRPSTO,LRPAT,COMPO+3,'RP!'
	mov RPL,R2      ;TPL
	mov RPH,R3      ;TPH
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R2,A        ;TPL
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R3,A
	ajmp RETURN


; SP@   ( -- a )
;       Push the current data stack pointer.

	@CODE SPAT,LSPAT,LRPSTO,3,'SP@'
	mov A,R3        ;TPH
	movx @R0,A      ;SPL
	dec R0          ;SPL
	mov A,R2        ;TPL
	movx @R0,A
	dec R0          ;SPL
	mov TPL,R0      ;SPL
	mov R3,SPH      ;TPH
	ajmp RETURN


; SP!   ( a -- )
;       Set the data stack pointer.

	@CODE SPSTO,LSPSTO,LSPAT,3,'SP!'
	mov SPL,R2      ;TPL
	mov SPH,R3      ;TPH
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R2,A        ;TPL
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R3,A        ;TPH
	ajmp RETURN


; DUP   ( w -- w w )
;       Duplicate the top stack item.

	@CODE DUPP,LDUPP,LSPSTO,3,'DUP'
	mov A,R3        ;TPH
	movx @R0,A      ;SPL
	dec R0          ;SPL
	mov A,R2        ;TPL
	movx @R0,A      ;SPL
	dec R0          ;SPL
	ajmp RETURN


; DROP  ( w -- )
;       Discard top stack item.

	@CODE DROP,LDROP,LDUPP,4,'DROP'
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R2,A        ;TPL
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R3,A        ;TPH
	ajmp RETURN


; SWAP  ( w1 w2 -- w2 w1 )
;       Exchange top two stack items.

	@CODE SWAP,LSWAP,LDROP,4,'SWAP'
	mov DPL,R0      ;SPL
	mov DPH,SPH
	inc DPL
	movx A,@DPTR
	xch A,R2        ;TPL
	movx @DPTR,A
	inc DPL
	movx A,@DPTR
	xch A,R3        ;TPH
	movx @DPTR,A
	ajmp RETURN


; FLIP  ( -- )
;       Swap rsp. exchange the two bytes on TOS

	@CODE FLIP,LFLIP,LSWAP,4,'FLIP'
	mov DPL,R2      ;TPL
	mov DPH,R3      ;TPH
	mov R2,DPH      ;TPL
	mov R3,DPL      ;TPH
	ajmp RETURN


; SPLIT  ( w -- wHI wLO)
;       splits the word in TOS into two words,
;       each containing one byte with leading zeroes

	@CODE SPLIT,LSPLIT,LFLIP,5,'SPLIT'

	clr A
	movx @R0,A      ;SPL
	dec R0          ;SPL
	mov A,R3        ;TPH
	movx @R0,A      ;SPL  Hi TOS Byte to
	dec R0          ;     lower byte of second stack element
	clr A
	mov R3,A        ;R2 remains unchanged
	ajmp RETURN


;JOIN   ( wHI wLO -- w )
;       Joins the lower bytes of the two stack top items
;       to one word held in TOS

	@CODE JOIN,LJOIN,LSPLIT,4,'JOIN'

	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R3,A        ;TPH
	inc R0          ;SPL
	ajmp RETURN


; OVER  ( w1 w2 -- w1 w2 w1 )
;       Copy second stack item to top.

	@CODE OVER,LOVER,LJOIN,4,'OVER'
	mov DPL,R0      ;SPL
	mov DPH,SPH
	mov A,R3        ;TPH
	movx @R0,A      ;SPL
	dec R0          ;SPL
	mov A,R2        ;TPL
	movx @R0,A      ;SPL
	dec R0          ;SPL
	inc DPL
	movx A,@DPTR
	mov R2,A        ;TPL
	inc DPL
	movx A,@DPTR
	mov R3,A        ;TPH
	ajmp RETURN


; 0<    ( n -- t )
;       Return true if n is negative.

	@CODE ZLESS,LZLESS,LOVER,2,'0<'
	mov A,R3        ;TPH
	jnb ACC.7,ZLESS1
	mov A,#$FF
	sjmp ZLESS2
ZLESS1: clr A
ZLESS2: mov R2,A        ;TPL
	mov R3,A        ;TPH
	ajmp RETURN


; AND   ( w w -- w )
;       Bitwise AND.

	@CODE ANDD,LANDD,LZLESS,3,'AND'
	inc r0          ;SPL
	movx A,@R0      ;SPL
	anl TPL,A
	inc R0          ;SPL
	movx A,@R0      ;SPL
	anl TPH,A
	ajmp RETURN


; OR    ( w w -- w )
;       Bitwise inclusive OR.

	@CODE ORR,LORR,LANDD,2,'OR'
	inc R0          ;SPL
	movx A,@R0      ;SPL
	orl TPL,A
	inc R0          ;SPL
	movx A,@R0      ;SPL
	orl TPH,A
	ajmp RETURN


; XOR   ( w w -- w )
;       Bitwise exclusive OR.

	@CODE XORR,LXORR,LORR,3,'XOR'
	inc R0          ;SPL
	movx A,@R0      ;SPL
	xrl TPL,A
	inc R0          ;SPL
	movx A,@R0      ;SPL
	xrl TPH,A
	ajmp RETURN


; UM+   ( w w -- w cy )
;       Add two numbers, return the sum and carry flag.

	@CODE UPLUS,LUPLUS,LXORR,3,'UM+'
	inc R0          ;SPL
	movx A,@R0      ;SPL
	add A,R2        ;TPL
	movx @R0,A      ;SPL
	inc R0          ;SPL
	movx A,@R0      ;SPL
	addc A,R3       ;TPH
	movx @R0,A      ;SPL
	dec R0          ;SPL
	dec R0          ;SPL
	clr A
	mov R3,A        ;TPH
	addc A,R3       ;TPH
	mov R2,A        ;TPL
	ajmp RETURN


;; Device dependent I/O

; !IO   ( -- )
;       Initialize the serial I/O devices.

	@CODE STOIO,LSTOIO,LUPLUS,3,'!IO'
	clr IE.4
	mov SCON,#$52
	ajmp RETURN



; ?RX   ( -- c T | F )
;       Return input character and true, or a false if no input.

	@CODE QRX,LQRX,LSTOIO,3,'?RX'
	mov A,R3        ;TPH
	movx @R0,A      ;SPL
	dec R0          ;SPL
	mov A,R2        ;TPL
	movx @R0,A      ;SPL
	dec R0          ;SPL
	jnb SCON.0,QRX1
	clr SCON.0
	mov R2,SBUF     ;TPL
	mov R3,#0       ;TPH
	mov A,R3        ;TPH
	movx @R0,A      ;SPL
	dec R0          ;SPL
	mov A,R2        ;TPL
	movx @R0,A      ;SPL
	dec R0          ;SPL
	mov R2,#$FF     ;TPL
	mov R3,#$FF     ;TPH
	ajmp RETURN
QRX1:  mov R2,#0        ;TPL
	mov R3,#0       ;TPH
	ajmp RETURN


; TX!   ( c -- )
;       Send character c to the output device.

	@CODE TXSTO,LTXSTO,LQRX,3,'TX!'
	jnb SCON.1,TXSTO1
	clr SCON.1
TXSTO1: mov SBUF,R2     ;TPL
TXSTO2: jnb SCON.1,TXSTO2
	clr SCON.1
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R2,A        ;TPL
	inc R0          ;SPL
	movx A,@R0      ;SPL
	mov R3,A        ;TPH
	ajmp RETURN


;; System and user variables

; doVAR ( -- a )
;       Run time routine for VARIABLE and CREATE.

	@COLON  DOVAR,LDOVAR,LTXSTO,COMPO+5,'doVAR'
	.DW  RFROM,EXIT


; UP    ( -- a )
;       Pointer to the user area.

	@COLON  UP,LUP,LDOVAR,2,'UP'
	.DW  DOVAR
	.DW  UPP


; doUSER ( -- a )
;       Run time routine for user variables.

	@COLON  DOUSE,LDOUSE,LUP,COMPO+6,'DOUSER'
	.DW  RFROM,AT,UP,AT,UPLUS,DROP,EXIT
			;Code changed because PLUS isn't available here
			;without forward references(WS)


; MACRO for creating user variables in the initial system

	.MACRO  @USER  LABEL,LLABEL,LINK,LEX,NAME,OFFSET
		@COLON  LABEL,LLABEL,LINK,LEX,NAME
		.DW  DOUSE,OFFSET
	.ENDM


; SP0   ( -- a )
;       Pointer to bottom of the data stack.

	@USER  SZERO,LSZERO,LDOUSE,3,'SP0',USZERO-UPP


; RP0   ( -- a )
;       Pointer to bottom of the return stack.

	@USER  RZERO,LRZERO,LSZERO,3,'RP0',URZERO-UPP


; '?KEY ( -- a )
;       Execution vector of ?KEY.

	@USER  TQKEY,LTQKEY,LRZERO,5,"'?KEY",UTQKEY-UPP


; 'EMIT ( -- a )
;       Execution vector of EMIT.

	@USER  TEMIT,LTEMIT,LTQKEY,5,"'EMIT",UTEMIT-UPP


; 'EXPECT  ( -- a )
;       Execution vector of EXPECT.

	@USER  TEXPE,LTEXPE,LTEMIT,7,"'EXPECT",UTEXPE-UPP


; 'TAP  ( -- a )
;       Execution vector of TAP.

	@USER  TTAP,LTTAP,LTEXPE,4,"'TAP",UTTAP-UPP


; 'ECHO ( -- a )
;       Execution vector of ECHO.

	@USER  TECHO,LTECHO,LTTAP,5,"'ECHO",UTECHO-UPP


; 'PROMPT ( -- a )
;       Execution vector of PROMPT.

	@USER  TPROM,LTPROM,LTECHO,7,"'PROMPT",UTPROM-UPP


; BASE  ( -- a )
;       Storage of the radix base for numeric I/O.

	@USER  BASE,LBASE,LTPROM,4,'BASE',UBASE-UPP


; tmp   ( -- a )
;       A temporary storage location used in parse and find.

	@USER  TEMP,LTEMP,LBASE,COMPO+3,'tmp',UTEMP-UPP


; SPAN  ( -- a )
;       Hold character count received by EXPECT.

	@USER  SPAN,LSPAN,LTEMP,4,'SPAN',USPAN-UPP


; >IN   ( -- a )
;       Hold the character pointer while parsing input stream.

	@USER  INN,LINN,LSPAN,3,'>IN',UINN-UPP


; #TIB  ( -- a )
;       Hold the current count and address of the terminal input buffer.

	@USER  NTIB,LNTIB,LINN,4,'#TIB',UNTIB-UPP


; CSP   ( -- a )
;       Hold the stack pointer for error checking.

	@USER  CSP,LCSP,LNTIB,3,'CSP',UCSP-UPP


; 'EVAL ( -- a )
;       Execution vector of EVAL.

	@USER  TEVAL,LTEVAL,LCSP,5,"'EVAL",UTEVAL-UPP


; 'NUMBER ( -- a )
;       Execution vector of NUMBER?.

	@USER  TNUMB,LTNUMB,LTEVAL,7,"'NUMBER",UTNUMB-UPP


; HLD   ( -- a )
;       Hold a pointer in building a numeric output string.

	@USER  HLD,LHLD,LTNUMB,3,'HLD',UHLD-UPP


; HANDLER ( -- a )
;       Hold the return stack pointer for error handling.

	@USER  HANDL,LHANDL,LHLD,7,'HANDLER',UHANDL-UPP


; VARAD ( -- a)
;        holds the appropriate PFA for global variable
;        to be defined next in an extra RAM portion

	@USER   VARAD,LVARAD,LHANDL,5,'VARAD',UVARAD-UPP


; USOFF ( -- a)
;        holds the appropriate offset for user variable
;        to be defined next

	@USER   USOFF,LUSOFF,LVARAD,5,'USOFF',UUSOFF-UPP


; CONTEXT ( -- a )
;       A area to specify vocabulary search order.

	@USER  CNTXT,LCNTXT,LUSOFF,7,'CONTEXT',UCNTXT-UPP


; CURRENT ( -- a )
;       Point to the vocabulary to be extended.

	@USER  CRRNT,LCRRNT,LCNTXT,7,'CURRENT',UCRRNT-UPP


; CP    ( -- a )
;       Point to the top of the dictionary.

	@USER   CP,LCP,LCRRNT,2,'CP',UCP-UPP

;       (WS): in the original eForth CP works as a specific code pointer.
;             In this Version, CP represents both code and name pointer
;             because headers and code are interleaved, both growing up.


; NP    ( -- a )
;       Point to the bottom of the name dictionary.

	@USER   NP,LNP,LCP,2,'NP',UNP-UPP

;       This USER variable is without funtion in this implementation,
;       but left for better compatibilitiy with the original eForth


; LAST  ( -- a )
;       Point to the last name in the name dictionary.

	@USER   LAST,LLAST,LNP,4,'LAST',ULAST-UPP


; forth ( -- a )
;       Point to the last name in the name dictionary.

	@USER   VFRTH,LVFRTH,LLAST,5,'forth',UVFRTH-UPP


;; Common functions

; FORTH ( -- )
;       Make FORTH the context vocabulary.

	@COLON  FORTH,LFORTH,LVFRTH,5,'FORTH'
	.DW  VFRTH,CNTXT,STORE,EXIT


; ?DUP  ( w -- w w | 0 )
;       Dup tos if its is not zero.

	@COLON  QDUP,LQDUP,LFORTH,4,'?DUP'
	.DW  DUPP
	.DW  QBRAN,QDUP1
	.DW  DUPP
QDUP1:  .DW  EXIT


; ROT   ( w1 w2 w3 -- w2 w3 w1 )
;       Rot 3rd item to top.

	@COLON  ROT,LROT,LQDUP,3,'ROT'
	.DW  TOR,SWAP,RFROM,SWAP,EXIT


; 2DROP ( w w -- )
;       Discard two items on stack.

	@COLON  DDROP,LDDROP,LROT,5,'2DROP'
	.DW  DROP,DROP,EXIT


; 2DUP  ( w1 w2 -- w1 w2 w1 w2 )
;       Duplicate top two items.

	@COLON  DDUP,LDDUP,LDDROP,4,'2DUP'
	.DW  OVER,OVER,EXIT


; +     ( w w -- sum )
;       Add top two items.

	@COLON  PLUS,LPLUS,LDDUP,1,'+'
	.DW  UPLUS,DROP,EXIT


; D+    ( d d -- d )
;       Double addition, as an example using UM+.
;
;       @COLON  DDPLUS,LDDPLUS,LPLUS,2,'D+'
;       .DW  TOR,SWAP,TOR,UPLUS
;       .DW  RFROM,RFROM,PLUS,PLUS,EXIT


; NOT   ( w -- w )
;       One's complement of tos.

	@COLON  INVER,LINVER,LPLUS,3,'NOT'
	.DW  DOLIT,-1,XORR,EXIT


; NEGATE ( n -- -n )
;       Two's complement of tos.

	@COLON  NEGAT,LNEGAT,LINVER,6,'NEGATE'
	.DW  INVER,DOLIT,1,PLUS,EXIT


; DNEGATE ( d -- -d )
;       Two's complement of top double.

	@COLON  DNEGA,LDNEGA,LNEGAT,7,'DNEGATE'
	.DW  INVER,TOR,INVER
	.DW  DOLIT,1,UPLUS
	.DW  RFROM,PLUS,EXIT


; -     ( n1 n2 -- n1-n2 )
;       Subtraction.

	@COLON  SUBB,LSUBB,LDNEGA,1,'-'
	.DW  NEGAT,PLUS,EXIT


; ABS   ( n -- n )
;       Return the absolute value of n.

	@COLON  ABSS,LABSS,LSUBB,3,'ABS'
	.DW  DUPP,ZLESS
	.DW  QBRAN,ABS1
	.DW  NEGAT
ABS1:   .DW  EXIT


; =     ( w w -- t )
;       Return true if top two are equal.

	@COLON  EQUAL,LEQUAL,LABSS,1,'='
	.DW  XORR
	.DW  QBRAN,EQU1
	.DW  DOLIT,0,EXIT        ;false flag
EQU1:   .DW  DOLIT,-1,EXIT       ;true flag


; U<    ( u u -- t )
;       Unsigned compare of top two items.

	@COLON  ULESS,LULESS,LEQUAL,2,'U<'
	.DW  DDUP,XORR,ZLESS
	.DW  QBRAN,ULES1
	.DW  SWAP,DROP,ZLESS,EXIT
ULES1:  .DW  SUBB,ZLESS,EXIT


; <     ( n1 n2 -- t )
;       Signed compare of top two items.

	@COLON  LESS,LLESS,LULESS,1,'<'
	.DW  DDUP,XORR,ZLESS
	.DW  QBRAN,LESS1
	.DW  DROP,ZLESS,EXIT
LESS1:  .DW  SUBB,ZLESS,EXIT


; MAX   ( n n -- n )
;       Return the greater of two top stack items.

	@COLON  MAX,LMAX,LLESS,3,'MAX'
	.DW  DDUP,LESS
	.DW  QBRAN,MAX1
	.DW  SWAP
MAX1:   .DW  DROP,EXIT


; MIN   ( n n -- n )
;       Return the smaller of top two stack items.

	@COLON  MIN,LMIN,LMAX,3,'MIN'
	.DW  DDUP,SWAP,LESS
	.DW  QBRAN,MIN1
	.DW  SWAP
MIN1:   .DW  DROP,EXIT


; WITHIN ( u ul uh -- t )
;       Return true if u is within the range of ul and uh.

	@COLON  WITHI,LWITHI,LMIN,6,'WITHIN'
	.DW  OVER,SUBB,TOR           ;ul <= u < uh
	.DW  SUBB,RFROM,ULESS,EXIT


;; Divide

; UM/MOD ( udl udh u -- ur uq )
;       Unsigned divide of a double by a single. Return mod and quotient.

	@COLON  UMMOD,LUMMOD,LWITHI,6,'UM/MOD'
	.DW  DDUP,ULESS
	.DW  QBRAN,UMM4
	.DW  NEGAT,DOLIT,15,TOR
UMM1:   .DW  TOR,DUPP,UPLUS
	.DW  TOR,TOR,DUPP,UPLUS
	.DW  RFROM,PLUS,DUPP
	.DW  RFROM,RAT,SWAP,TOR
	.DW  UPLUS,RFROM,ORR
	.DW  QBRAN,UMM2
	.DW  TOR,DROP,DOLIT,1,PLUS,RFROM
	.DW  BRAN,UMM3
UMM2:   .DW  DROP
UMM3:   .DW  RFROM
	.DW  DONXT,UMM1
	.DW  DROP,SWAP,EXIT
UMM4:   .DW  DROP,DDROP
	.DW  DOLIT,-1,DUPP,EXIT  ;overflow, return max


; M/MOD ( d n -- r q )
;       Signed floored divide of double by single. Return mod and quotient.

	@COLON  MSMOD,LMSMOD,LUMMOD,5,'M/MOD'
	.DW  DUPP,ZLESS,DUPP,TOR
	.DW  QBRAN,MMOD1
	.DW  NEGAT,TOR,DNEGA,RFROM
MMOD1:  .DW  TOR,DUPP,ZLESS
	.DW  QBRAN,MMOD2
	.DW  RAT,PLUS
MMOD2:  .DW  RFROM,UMMOD,RFROM
	.DW  QBRAN,MMOD3
	.DW  SWAP,NEGAT,SWAP
MMOD3:  .DW  EXIT


; /MOD  ( n n -- r q )
;       Signed divide. Return mod and quotient.

	@COLON  SLMOD,LSLMOD,LMSMOD,4,'/MOD'
	.DW  OVER,ZLESS,SWAP,MSMOD,EXIT


; MOD   ( n n -- r )
;       Signed divide. Return mod only.

	@COLON  MODD,LMODD,LSLMOD,3,'MOD'
	.DW  SLMOD,DROP,EXIT


; /     ( n n -- q )
;       Signed divide. Return quotient only.

	@COLON  SLASH,LSLASH,LMODD,1,'/'
	.DW  SLMOD,SWAP,DROP,EXIT


;; Multiply

; UM*   ( u u -- ud )
;       Unsigned multiply. Return double product.

	@COLON  UMSTA,LUMSTA,LSLASH,3,'UM*'
	.DW  DOLIT,0,SWAP,DOLIT,15,TOR
UMST1:  .DW  DUPP,UPLUS,TOR,TOR
	.DW  DUPP,UPLUS,RFROM,PLUS,RFROM
	.DW  QBRAN,UMST2
	.DW  TOR,OVER,UPLUS,RFROM,PLUS
UMST2:  .DW  DONXT,UMST1
	.DW  ROT,DROP,EXIT


; *     ( n n -- n )
;       Signed multiply. Return single product.

	@COLON  STAR,LSTAR,LUMSTA,1,'*'
	.DW  UMSTA,DROP,EXIT


; M*    ( n n -- d )
;       Signed multiply. Return double product.

	@COLON  MSTAR,LMSTAR,LSTAR,2,'M*'
	.DW  DDUP,XORR,ZLESS,TOR
	.DW  ABSS,SWAP,ABSS,UMSTA
	.DW  RFROM
	.DW  QBRAN,MSTA1
	.DW  DNEGA
MSTA1:  .DW  EXIT


; */MOD ( n1 n2 n3 -- r q )
;       Multiply n1 and n2, then divide by n3. Return mod and quotient.

	@COLON  SSMOD,LSSMOD,LMSTAR,5,'*/MOD'
	.DW  TOR,MSTAR,RFROM,MSMOD,EXIT


; */    ( n1 n2 n3 -- q )
;       Multiply n1 by n2, then divide by n3. Return quotient only.

	@COLON  STASL,LSTASL,LSSMOD,2,'*/'
	.DW  SSMOD,SWAP,DROP,EXIT


;; Miscellaneous

; CELL+ ( a -- a )
;       Add cell size in byte to address.

	@COLON  CELLP,LCELLP,LSTASL,5,'CELL+'
	.DW  DOLIT,CELLL,PLUS,EXIT


; CELL- ( a -- a )
;       Subtract cell size in byte from address.

	@COLON  CELLM,LCELLM,LCELLP,5,'CELL-'
	.DW  DOLIT,0-CELLL,PLUS,EXIT


; CELLS ( n -- n )
;       Multiply tos by cell size in bytes.

	@COLON  CELLS,LCELLS,LCELLM,5,'CELLS'
	.DW  DOLIT,CELLL,STAR,EXIT


; ALIGNED ( b -- a )
;       Align address to the cell boundary.

	@COLON  ALGND,LALGND,LCELLS,7,'ALIGNED'
	.DW  DUPP,DOLIT,0,DOLIT,CELLL
	.DW  UMMOD,DROP,DUPP
	.DW  QBRAN,ALGN1
	.DW  DOLIT,CELLL,SWAP,SUBB
ALGN1:  .DW  PLUS,EXIT


; BL    ( -- 32 )
;       Return 32, the blank character.

	@COLON  BLANK,LBLANK,LALGND,2,'BL'
	.DW  DOLIT,' ',EXIT


; >CHAR ( c -- c )
;       Filter non-printing characters.

	@COLON  TCHAR,LTCHAR,LBLANK,5,'>CHAR'
	.DW  DOLIT,07FH,ANDD,DUPP    ;mask msb
	.DW  DOLIT,127,BLANK,WITHI   ;check for printable
	.DW  QBRAN,TCHA1
	.DW  DROP,DOLIT,'_'          ;replace non-printables
TCHA1:  .DW  EXIT


; DEPTH ( -- n )
;       Return the depth of the data stack.

	@COLON  DEPTH,LDEPTH,LTCHAR,5,'DEPTH'
	.DW  SPAT,SZERO,AT,SWAP,SUBB
	.DW  DOLIT,CELLL,SLASH,EXIT


; PICK  ( ... +n -- ... w )
;       Copy the nth stack item to tos.

	@COLON  PICK,LPICK,LDEPTH,4,'PICK'
	.DW  DOLIT,1,PLUS,CELLS
	.DW  DOLIT,1,PLUS
	.DW  SPAT,PLUS,AT,EXIT


;; Memory access

; +!    ( n a -- )
;       Add n to the contents at address a.

	@COLON  PSTOR,LPSTOR,LPICK,2,'+!'
	.DW  SWAP,OVER,AT,PLUS
	.DW  SWAP,STORE,EXIT


; 2!    ( d a -- )
;       Store the double integer to address a.

	@COLON  DSTOR,LDSTOR,LPSTOR,2,'2!'
	.DW  SWAP,OVER,STORE
	.DW  CELLP,STORE,EXIT


; 2@    ( a -- d )
;       Fetch double integer from address a.

	@COLON  DAT,LDAT,LDSTOR,2,'2@'
	.DW  DUPP,CELLP,AT
	.DW  SWAP,AT,EXIT


; COUNT ( b -- b +n )
;       Return count byte of a string and add 1 to byte address.

	@COLON  COUNT,LCOUNT,LDAT,5,'COUNT'
	.DW  DUPP,DOLIT,1,PLUS
	.DW  SWAP,CAT,EXIT


; HERE  ( -- a )
;       Return the top of the dictionary.(Headers & code interleaved)

	@COLON  HERE,LHERE,LCOUNT,4,'HERE'
	.DW  CP,AT,EXIT


; PAD   ( -- a )
;       Return the address of a temporary buffer.

	@COLON  PAD,LPAD,LHERE,3,'PAD'
	.DW  HERE,DOLIT,80,PLUS,EXIT


; TIB   ( -- a )
;       Return the address of the terminal input buffer.

	@COLON  TIB,LTIB,LPAD,3,'TIB'
	.DW  NTIB,CELLP,AT,EXIT


; @EXECUTE ( a -- )
;       Execute vector stored in address a.

	@COLON  ATEXE,LATEXE,LTIB,8,'@EXECUTE'
	.DW  AT,QDUP         ;?address or zero
	.DW  QBRAN,EXE1
	.DW  EXECU           ;execute if non-zero
EXE1:   .DW  EXIT            ;do nothing if zero


; CMOVE ( b1 b2 u -- )
;       Copy u bytes from b1 to b2.

	@COLON  CMOVE,LCMOVE,LATEXE,5,'CMOVE'
	.DW  TOR
	.DW  BRAN,CMOV2
CMOV1:  .DW  TOR,DUPP,CAT
	.DW  RAT,CSTOR
	.DW  DOLIT,1,PLUS
	.DW  RFROM,DOLIT,1,PLUS
CMOV2:  .DW  DONXT,CMOV1
	.DW  DDROP,EXIT


; FILL  ( b u c -- )
;       Fill u bytes of character c to area beginning at b.

	@COLON  FILL,LFILL,LCMOVE,4,'FILL'
	.DW  SWAP,TOR,SWAP
	.DW  BRAN,FILL2
FILL1:  .DW  DDUP,CSTOR,DOLIT,1,PLUS
FILL2:  .DW  DONXT,FILL1
	.DW  DDROP,EXIT


; -TRAILING ( b u -- b u )
;       Adjust the count to eliminate trailing white space.

	@COLON  DTRAI,LDTRAI,LFILL,9,'-TRAILING'
	.DW  TOR
	.DW  BRAN,DTRA2
DTRA1:  .DW  BLANK,OVER,RAT,PLUS,CAT,LESS
	.DW  QBRAN,DTRA2
	.DW  RFROM,DOLIT,1,PLUS,EXIT ;adjusted count
DTRA2:  .DW  DONXT,DTRA1
	.DW  DOLIT,0,EXIT        ;count=0


; PACK$ ( b u a -- a )
;       Build a counted string with u characters from b. Null fill.

	@COLON  PACKS,LPACKS,LDTRAI,5,'PACK$'
	.DW  ALGND,DUPP,TOR      ;strings only on cell boundary
	.DW  OVER,DUPP,DOLIT,0
	.DW  DOLIT,CELLL,UMMOD,DROP  ;count mod cell
	.DW  SUBB,OVER,PLUS
	.DW  DOLIT,0,SWAP,STORE  ;null fill cell
	.DW  DDUP,CSTOR,DOLIT,1,PLUS ;save count
	.DW  SWAP,CMOVE,RFROM,EXIT   ;move string


;; Numeric output, single precision

; DIGIT ( u -- c )
;       Convert digit u to a character.

	@COLON  DIGIT,LDIGIT,LPACKS,5,'DIGIT'
	.DW  DOLIT,9,OVER,LESS
	.DW  DOLIT,7,ANDD,PLUS
	.DW  DOLIT,'0',PLUS,EXIT


; EXTRACT ( n base -- n c )
;       Extract the least significant digit from n.

	@COLON  EXTRC,LEXTRC,LDIGIT,7,'EXTRACT'
	.DW  DOLIT,0,SWAP,UMMOD
	.DW  SWAP,DIGIT,EXIT


; <#    ( -- )
;       Initiate the numeric output process.

	@COLON  BDIGS,LBDIGS,LEXTRC,2,'<#'
	.DW  PAD,HLD,STORE,EXIT


; HOLD  ( c -- )
;       Insert a character into the numeric output string.

	@COLON  HOLD,LHOLD,LBDIGS,4,'HOLD'
	.DW  HLD,AT,DOLIT,1,SUBB
	.DW  DUPP,HLD,STORE,CSTOR,EXIT


; #     ( u -- u )
;       Extract one digit from u and append the digit to output string.

	@COLON  DIG,LDIG,LHOLD,1,'#'
	.DW  BASE,AT,EXTRC,HOLD,EXIT


; #S    ( u -- 0 )
;       Convert u until all digits are added to the output string.

	@COLON  DIGS,LDIGS,LDIG,2,'#S'
DIGS1:   .DW  DIG,DUPP
	.DW  QBRAN,DIGS2
	.DW  BRAN,DIGS1
DIGS2:  .DW  EXIT


; SIGN  ( n -- )
;       Add a minus sign to the numeric output string.

	@COLON  SIGN,LSIGN,LDIGS,4,'SIGN'
	.DW  ZLESS
	.DW  QBRAN,SIGN1
	.DW  DOLIT,'-',HOLD
SIGN1:  .DW  EXIT


; #>    ( w -- b u )
;		Prepare the output string to be TYPE'd.

	@COLON  EDIGS,LEDIGS,LSIGN,2,'#>'
	.DW  DROP,HLD,AT
	.DW  PAD,OVER,SUBB,EXIT


; str   ( n -- b u )
;       Convert a signed integer to a numeric string.

	@COLON  STR,LSTR,LEDIGS,3,'str'
	.DW  DUPP,TOR,ABSS
	.DW  BDIGS,DIGS,RFROM
	.DW  SIGN,EDIGS,EXIT


; HEX   ( -- )
;       Use radix 16 as base for numeric conversions.

	@COLON  HEX,LHEX,LSTR,3,'HEX'
	.DW  DOLIT,16,BASE,STORE,EXIT


; DECIMAL ( -- )
;       Use radix 10 as base for numeric conversions.

	@COLON  DECIM,LDECIM,LHEX,7,'DECIMAL'
	.DW  DOLIT,10,BASE,STORE,EXIT


;; Numeric input, single precision

; DIGIT? ( c base -- u t )
;       Convert a character to its numeric value. A flag indicates success.

	@COLON  DIGTQ,LDIGTQ,LDECIM,6,'DIGIT?'
	.DW  TOR,DOLIT,'0',SUBB
	.DW  DOLIT,9,OVER,LESS
	.DW  QBRAN,DGTQ1
	.DW  DOLIT,7,SUBB
	.DW  DUPP,DOLIT,10,LESS,ORR
DGTQ1:  .DW  DUPP,RFROM,ULESS,EXIT


; NUMBER? ( a -- n T | a F )
;       Convert a number string to integer. Push a flag on tos.

	@COLON  NUMBQ,LNUMBQ,LDIGTQ,7,'NUMBER?'
	.DW  BASE,AT,TOR,DOLIT,0,OVER,COUNT
	.DW  OVER,CAT,DOLIT,'$',EQUAL
	.DW  QBRAN,NUMQ1
	.DW  HEX,SWAP,DOLIT,1,PLUS
	.DW  SWAP,DOLIT,1,SUBB
NUMQ1:  .DW  OVER,CAT,DOLIT,'-',EQUAL,TOR
	.DW  SWAP,RAT,SUBB,SWAP,RAT,PLUS,QDUP
	.DW  QBRAN,NUMQ6
	.DW  DOLIT,1,SUBB,TOR
NUMQ2:  .DW  DUPP,TOR,CAT,BASE,AT,DIGTQ
	.DW  QBRAN,NUMQ4
	.DW  SWAP,BASE,AT,STAR,PLUS,RFROM
	.DW  DOLIT,1,PLUS
	.DW  DONXT,NUMQ2
	.DW  RAT,SWAP,DROP
	.DW  QBRAN,NUMQ3
	.DW  NEGAT
NUMQ3:  .DW  SWAP
	.DW  BRAN,NUMQ5
NUMQ4:  .DW  RFROM,RFROM,DDROP,DDROP,DOLIT,0
NUMQ5:  .DW  DUPP
NUMQ6:  .DW  RFROM,DDROP
	.DW  RFROM,BASE,STORE,EXIT


;; Basic I/O

; ?KEY  ( -- c T | F )
;       Return input character and true, or a false if no input.

	@COLON  QKEY,LQKEY,LNUMBQ,4,'?KEY'
	.DW  TQKEY,ATEXE,EXIT


; KEY   ( -- c )
;       Wait for and return an input character.

	@COLON  KEY,LKEY,LQKEY,3,'KEY'
KEY1:   .DW  QKEY
	.DW  QBRAN,KEY1
	.DW  EXIT


; EMIT  ( c -- )
;       Send a character to the output device.

	@COLON  EMIT,LEMIT,LKEY,4,'EMIT'
	.DW  TEMIT,ATEXE,EXIT


; NUF?  ( -- t )
;       Return false if no input, else pause and if CR return true.

	@COLON  NUFQ,LNUFQ,LEMIT,4,'NUF?'
	.DW  QKEY,DUPP
	.DW  QBRAN,NUFQ1
	.DW  DDROP,KEY,DOLIT,CRR,EQUAL
NUFQ1:  .DW  EXIT


; PACE  ( -- )
;       Send a pace character for the file downloading process.

	 @COLON  PACE,LPACE,LNUFQ,4,'PACE'
	 .DW  DOLIT,11,EMIT,EXIT


; SPACE  ( -- )
;        Send the blank character to the output device.

	@COLON  SPACE,LSPACE,LPACE,5,'SPACE'
	.DW  BLANK,EMIT,EXIT


; SPACES ( +n -- )
;       Send n spaces to the output device.

	@COLON  SPACS,LSPACS,LSPACE,6,'SPACES'
	.DW  DOLIT,0,MAX,TOR
	.DW  BRAN,CHAR2
CHAR1:  .DW  SPACE
CHAR2:  .DW  DONXT,CHAR1
	.DW  EXIT


; TYPE  ( b u -- )
;       Output u characters from b.

	@COLON  TYPEE,LTYPEE,LSPACS,4,'TYPE'
	.DW  TOR
	.DW  BRAN,TYPE2
TYPE1:  .DW  DUPP,CAT,EMIT
	.DW  DOLIT,1,PLUS
TYPE2:  .DW  DONXT,TYPE1
	.DW  DROP,EXIT


; CR    ( -- )
;       Output a carriage return and a line feed.

	@COLON  CR,LCR,LTYPEE,2,'CR'
	.DW  DOLIT,CRR,EMIT
	.DW  DOLIT,LF,EMIT,EXIT


; do$   ( -- a )
;       Return the address of a compiled string.

	@COLON  DOSTR,LDOSTR,LCR,COMPO+3,'do$'
	.DW  RFROM,RAT,RFROM,COUNT,PLUS
	.DW  ALGND,TOR,SWAP,TOR,EXIT


; $"|   ( -- a )
;       Run time routine compiled by $". Return address of a compiled string.

	@COLON  STRQP,LSTRQP,LDOSTR,COMPO+3,'$"|'
	.DW  DOSTR,EXIT      ;force a call to do$


; ."|   ( -- )
;       Run time routine of ." . Output a compiled string.

	@COLON  DOTQP,LDOTQP,LSTRQP,COMPO+3,'."|'
	.DW  DOSTR,COUNT,TYPEE,EXIT


; .R    ( n +n -- )
;       Display an integer in a field of n columns, right justified.

	@COLON  DOTR,LDOTR,LDOTQP,2,'.R'
	.DW  TOR,STR,RFROM,OVER,SUBB
	.DW  SPACS,TYPEE,EXIT


; U.R   ( u +n -- )
;       Display an unsigned integer in n column, right justified.

	@COLON  UDOTR,LUDOTR,LDOTR,3,'U.R'
	.DW  TOR,BDIGS,DIGS,EDIGS
	.DW  RFROM,OVER,SUBB
	.DW  SPACS,TYPEE,EXIT


; U.    ( u -- )
;       Display an unsigned integer in free format.

	@COLON  UDOT,LUDOT,LUDOTR,2,'U.'
	.DW  BDIGS,DIGS,EDIGS
	.DW  SPACE,TYPEE,EXIT


; .     ( w -- )
;       Display an integer in free format, preceeded by a space.

	@COLON  DOT,LDOT,LUDOT,1,'.'
	.DW  BASE,AT,DOLIT,10,XORR   ;?decimal
	.DW  QBRAN,DOT1
	.DW  UDOT,EXIT       ;no, display unsigned
DOT1:   .DW  STR,SPACE,TYPEE,EXIT    ;yes, display signed


; ?     ( a -- )
;       Display the contents in a memory cell.

	@COLON  QUEST,LQUEST,LDOT,1,'?'
	.DW  AT,DOT,EXIT


;; Parsing

; parse ( b u c -- b u delta ; <string> )
;       Scan string delimited by c. Return found string and its offset.

	@COLON  PARS,LPARS,LQUEST,5,'parse'
	.DW  TEMP,STORE,OVER,TOR,DUPP
	.DW  QBRAN,PARS8
	.DW  DOLIT,1,SUBB,TEMP,AT,BLANK,EQUAL
	.DW  QBRAN,PARS3
	.DW  TOR
PARS1:  .DW  BLANK,OVER,CAT      ;skip leading blanks ONLY
	.DW  SUBB,ZLESS,INVER
	.DW  QBRAN,PARS2
	.DW  DOLIT,1,PLUS
	.DW  DONXT,PARS1
	.DW  RFROM,DROP,DOLIT,0,DUPP,EXIT
PARS2:  .DW  RFROM
PARS3:  .DW  OVER,SWAP
	.DW  TOR
PARS4:  .DW  TEMP,AT,OVER,CAT,SUBB   ;scan for delimiter
	.DW  TEMP,AT,BLANK,EQUAL
	.DW  QBRAN,PARS5
	.DW  ZLESS
PARS5:  .DW  QBRAN,PARS6
	.DW  DOLIT,1,PLUS
	.DW  DONXT,PARS4
	.DW  DUPP,TOR
	.DW  BRAN,PARS7
PARS6:  .DW  RFROM,DROP,DUPP
	.DW  DOLIT,1,PLUS,TOR
PARS7:  .DW  OVER,SUBB
	.DW  RFROM,RFROM,SUBB,EXIT
PARS8:  .DW  OVER,RFROM,SUBB,EXIT


; PARSE ( c -- b u ; <string> )
;       Scan input stream and return counted string delimited by c.

	@COLON  PARSE,LPARSE,LPARS,5,'PARSE'
	.DW  TOR,TIB,INN,AT,PLUS ;current input buffer pointer
	.DW  NTIB,AT,INN,AT,SUBB ;remaining count
	.DW  RFROM,PARS,INN,PSTOR,EXIT


; .(    ( -- )
;       Output following string up to next ) .

	@COLON  DOTPR,LDOTPR,LPARSE,IMEDD+2,'.('
	.DW  DOLIT,')',PARSE,TYPEE,EXIT


; (     ( -- )
;       Ignore following string up to next ) . A comment.

	@COLON  PAREN,LPAREN,LDOTPR,IMEDD+1,'('
	.DW  DOLIT,')',PARSE,DDROP,EXIT


; \     ( -- )
;       Ignore following text till the end of line.

	@COLON  BKSLA,LBKSLA,LPAREN,IMEDD+1,'\'
	.DW  NTIB,AT,INN,STORE,EXIT


; CHAR  ( -- c )
;       Parse next word and return its first character.

	@COLON  CHAR,LCHAR,LBKSLA,4,'CHAR'
	.DW  BLANK,PARSE,DROP,CAT,EXIT


; TOKEN ( -- a ; <string> )
;       Parse a word from input stream and copy it to name dictionary.

	@COLON  TOKEN,LTOKEN,LCHAR,5,'TOKEN'
	.DW     BLANK,PARSE,DOLIT,31,MIN
;        .DW     CP,AT  ;bisherige Testversion
	.DW     CP,AT,ALGND,CELLP,CELLP ;nchst fllige na
	.DW     PACKS
;        .DW    DUPP,CAT,test,DROP
	.DW     EXIT


; WORD  ( c -- a ; <string> )
;       Parse a word from input stream and copy it to code dictionary.

	@COLON  WORDD,LWORDD,LTOKEN,4,'WORD'
	.DW  PARSE,HERE,PACKS,EXIT


;; Dictionary search

; NAME> ( na -- ca )
;       Return a code address given a name address.

	@COLON  NAMET,LNAMET,LWORDD,5,'NAME>'
	.DW  CELLM,CELLM,AT,EXIT


; SAME? ( a a u -- a a f \ -0+ )
;       Compare u cells in two strings. Return 0 if identical.

	@COLON  SAMEQ,LSAMEQ,LNAMET,5,'SAME?'
	.DW  TOR
	.DW  BRAN,SAME2
SAME1:  .DW  OVER,RAT,CELLS,PLUS,AT
	.DW  OVER,RAT,CELLS,PLUS,AT
	.DW  SUBB,QDUP
	.DW  QBRAN,SAME2
	.DW  RFROM,DROP,EXIT     ;strings not equal
SAME2:  .DW  DONXT,SAME1
	.DW  DOLIT,0,EXIT        ;strings equal


; find  ( a va -- ca na | a F )
;       Search a vocabulary for a string. Return ca and na if succeeded.

	@COLON  FIND,LFIND,LSAMEQ,4,'find'
	.DW  SWAP,DUPP,CAT
	.DW  DOLIT,CELLL,SLASH,TEMP,STORE
	.DW  DUPP,AT,TOR,CELLP,SWAP
FIND1:  .DW  AT,DUPP
	.DW  QBRAN,FIND6
	.DW  DUPP,AT,DOLIT,MASKK,ANDD,RAT,XORR
	.DW  QBRAN,FIND2
	.DW  CELLP,DOLIT,-1      ;true flag
	.DW  BRAN,FIND3
FIND2:  .DW  CELLP,TEMP,AT,SAMEQ
FIND3:  .DW  BRAN,FIND4
FIND6:  .DW  RFROM,DROP
	.DW  SWAP,CELLM,SWAP,EXIT
FIND4:  .DW  QBRAN,FIND5
	.DW  CELLM,CELLM
	.DW  BRAN,FIND1
FIND5:  .DW  RFROM,DROP,SWAP,DROP
	.DW  CELLM
	.DW  DUPP,NAMET,SWAP,EXIT


; NAME? ( a -- ca na | a F )
;       Search all context vocabularies for a string.

	@COLON  NAMEQ,LNAMEQ,LFIND,5,'NAME?'
	.DW  CNTXT,DUPP,DAT,XORR ;?context=also
	.DW  QBRAN,NAMQ1
	.DW  CELLM           ;no, start with context
NAMQ1:  .DW  TOR
NAMQ2:  .DW  RFROM,CELLP,DUPP,TOR    ;next in search order
	.DW  AT,QDUP
	.DW  QBRAN,NAMQ3
	.DW  FIND,QDUP       ;search vocabulary
	.DW  QBRAN,NAMQ2
	.DW  RFROM,DROP,EXIT     ;found name
NAMQ3:  .DW  RFROM,DROP      ;name not found
	.DW  DOLIT,0,EXIT        ;false flag


;; Terminal response

; ^H    ( bot eot cur -- bot eot cur )
;       Backup the cursor by one character.

	@COLON  BKSP,LBKSP,LNAMEQ,2,'^H'
	.DW  TOR,OVER,RFROM,SWAP,OVER,XORR
	.DW  QBRAN,BACK1
	.DW  DOLIT,BKSPP,TECHO,ATEXE,DOLIT,1,SUBB
	.DW  BLANK,TECHO,ATEXE
	.DW  DOLIT,BKSPP,TECHO,ATEXE
BACK1:  .DW  EXIT


; TAP   ( bot eot cur c -- bot eot cur )
;       Accept and echo the key stroke and bump the cursor.

	@COLON  TAP,LTAP,LBKSP,3,'TAP'
	.DW  DUPP,TECHO,ATEXE
	.DW  OVER,CSTOR,DOLIT,1,PLUS,EXIT


; kTAP  ( bot eot cur c -- bot eot cur )
;       Process a key stroke, CR or backspace.

	@COLON  KTAP,LKTAP,LTAP,4,'kTAP'
	.DW  DUPP,DOLIT,CRR,XORR
	.DW  QBRAN,KTAP2
	.DW  DOLIT,BKSPP,XORR
	.DW  QBRAN,KTAP1
	.DW  BLANK,TAP,EXIT
KTAP1:  .DW  BKSP,EXIT
KTAP2:  .DW  DROP,SWAP,DROP,DUPP,EXIT


; accept ( b u -- b u )
;       Accept characters to input buffer. Return with actual count.

	@COLON  ACCEP,LACCEP,LKTAP,6,'accept'
	.DW  OVER,PLUS,OVER
ACCP1:  .DW  DDUP,XORR
	.DW  QBRAN,ACCP4
	.DW  KEY,DUPP
;       .DW  BLANK,SUBB,DOLIT,95,ULESS
	.DW  BLANK,DOLIT,127,WITHI
	.DW  QBRAN,ACCP2
	.DW  TAP
	.DW  BRAN,ACCP3
ACCP2:  .DW  TTAP,ATEXE
ACCP3:  .DW  BRAN,ACCP1
ACCP4:  .DW  DROP,OVER,SUBB,EXIT


; EXPECT ( b u -- )
;       Accept input stream and store count in SPAN.

	@COLON  EXPEC,LEXPEC,LACCEP,6,'EXPECT'
	.DW  TEXPE,ATEXE,SPAN,STORE,DROP,EXIT


; QUERY ( -- )
;       Accept input stream to terminal input buffer.

	@COLON  QUERY,LQUERY,LEXPEC,5,'QUERY'
	.DW  TIB,DOLIT,80,TEXPE,ATEXE,NTIB,STORE
	.DW  DROP,DOLIT,0,INN,STORE,EXIT


;; Error handling

; CATCH ( ca -- 0 | err# )
;       Execute word at ca and set up an error frame for it.

	@COLON  CATCH,LCATCH,LQUERY,5,'CATCH'
	.DW  SPAT,TOR,HANDL,AT,TOR   ;save error frame
	.DW  RPAT,HANDL,STORE,EXECU  ;execute
	.DW  RFROM,HANDL,STORE   ;restore error frame
	.DW  RFROM,DROP,DOLIT,0,EXIT ;no error


; THROW ( err# -- err# )
;       Reset system to current local error frame an update error flag.

	@COLON  THROW,LTHROW,LCATCH,5,'THROW'
	.DW  HANDL,AT,RPSTO      ;restore return stack
	.DW  RFROM,HANDL,STORE   ;restore handler frame
	.DW  RFROM,SWAP,TOR,SPSTO    ;restore data stack
	.DW  DROP,RFROM,EXIT


; NULL$ ( -- a )
;       Return address of a null string with zero count.

	@COLON  NULLS,LNULLS,LTHROW,5,'NULL$'
	.DW  DOVAR           ;emulate CREATE
	.DW  0
	.DB 99,111,121,111,116,101
	@ALIGN


; ABORT ( -- )
;       Reset data stack and jump to QUIT.

	@COLON  ABORT,LABORT,LNULLS,5,'ABORT'
	.DW  NULLS,THROW


; abort" ( f -- )
;       Run time routine of ABORT" . Abort with a message.

	@COLON  ABORQ,LABORQ,LABORT,COMPO+6,'abort"'
	.DW  QBRAN,ABOR1     ;text flag
	.DW  DOSTR,THROW     ;pass error string
ABOR1:  .DW  DOSTR,DROP,EXIT     ;drop error


;; The text interpreter

; $INTERPRET ( a -- )
;       Interpret a word. If failed, try to convert it to an integer.

	@COLON  INTER,LINTER,LABORQ,10,'$INTERPRET'
	.DW  NAMEQ,QDUP      ;?defined
	.DW  QBRAN,INTE1
	.DW  AT,DOLIT,COMPO,ANDD ;?compile only lexicon bits
	@DISTR  ABORQ,13,' compile only'
	.DW  EXECU,EXIT      ;execute defined word
INTE1:  .DW  TNUMB,ATEXE     ;convert a number
	.DW  QBRAN,INTE2
	.DW  EXIT
INTE2:  .DW  THROW           ;error


; [     ( -- )
;       Start the text interpreter.

	@COLON  LBRAC,LLBRAC,LINTER,IMEDD+1,'['
	.DW  DOLIT,INTER,TEVAL,STORE,EXIT


; .OK   ( -- )
;       Display 'ok' only while interpreting.

	@COLON  DOTOK,LDOTOK,LLBRAC,3,'.OK'
	.DW  DOLIT,INTER,TEVAL,AT,EQUAL
	.DW  QBRAN,DOTO1
	@DISTR  DOTQP,3,' ok'
DOTO1:  .DW  CR,EXIT


; ?STACK ( -- )
;       Abort if the data stack underflows.

	@COLON  QSTAC,LQSTAC,LDOTOK,6,'?STACK'
	.DW  DEPTH,DOLIT,$78
	.DW  SWAP,SUBB,ZLESS    ;check only for max 8 cells
				;underflow and near-overflow
	@DISTR  ABORQ,10,' underflow'
	.DW  EXIT


; EVAL  ( -- )
;       Interpret the input stream.

	@COLON  EVAL,LEVAL,LQSTAC,4,'EVAL'
EVAL1:  .DW  TOKEN,DUPP,CAT      ;?input stream empty
	.DW  QBRAN,EVAL2
	.DW  TEVAL,ATEXE,QSTAC   ;evaluate input, check stack
	.DW  BRAN,EVAL1
EVAL2:  .DW  DROP,TPROM,ATEXE,EXIT   ;prompt


;; Shell

; PRESET ( -- )
;        Reset data stack pointer and the terminal input buffer.

	@COLON  PRESE,LPRESE,LEVAL,6,'PRESET'
	.DW  SZERO,AT,SPSTO
	.DW  DOLIT,TIBB,NTIB,CELLP,STORE,EXIT


; xio   ( a a a -- )
;       Reset the I/O vectors 'EXPECT, 'TAP, 'ECHO and 'PROMPT.

	@COLON  XIO,LXIO,LPRESE,COMPO+3,'xio'
	.DW  DOLIT,ACCEP,TEXPE,DSTOR
	.DW  TECHO,DSTOR,EXIT


; FILE  ( -- )
;       Select I/O vectors for file download.

	@COLON FILE,LFILE,LXIO,4,'FILE'
	.DW  DOLIT,PACE,DOLIT,DROP
	.DW  DOLIT,KTAP,XIO,EXIT


; HAND  ( -- )
;       Select I/O vectors for terminal interface.

	@COLON  HAND,LHAND,LFILE,4,'HAND'
	.DW  DOLIT,DOTOK,DOLIT,EMIT
	.DW  DOLIT,KTAP,XIO,EXIT


; I/O   ( -- a )
;       Array to store default I/O vectors.

	@COLON  ISLO,LISLO,LHAND,3,'I/O'
	.DW  DOVAR           ;emulate CREATE
	.DW  QRX,TXSTO       ;default I/O vectors


; CONSOLE ( -- )
;       Initiate terminal interface.

	@COLON  CONSO,LCONSO,LISLO,7,'CONSOLE'
	.DW  ISLO,DAT,TQKEY,DSTOR    ;restore default I/O device
	.DW  HAND,EXIT       ;keyboard input


;       SAVEBIN  ( a u -- )
;           pushes u arbitrary bytes starting from address a
;           via Serial Com. Uses XON/XOFF protocol. Simple data
;           format without header nor security.

	@COLON  SAVBI,LSAVBI,LCONSO,7,'SAVEBIN'
	.DW  TOR,BRAN,SAVBI2
SAVBI1: .DW  DUPP,CAT,EMIT,DOLIT,1,PLUS
SAVBI2: .DW  DONXT,SAVBI1
	.DW  DROP,EXIT


;       LOADBIN  ( a u -- )
;           loads u arbitrary bytes to address a via SerialCom.
;           No semantic or allocation precautions, use only if you
;           are looking through.

	@COLON  LODBI,LLODBI,LSAVBI,7,'LOADBIN'
	.DW  PACE
	.DW  TOR,BRAN,LODBI2
LODBI1: .DW  DUPP,KEY,SWAP,CSTOR,DOLIT,1,PLUS
LODBI2: .DW  DONXT,LODBI1
	.DW  DROP,EXIT


;SAVESYSTEM ( -- )
;       Save the RAM portion of dictionary extension and the USER area
;       via SerialCom

	@COLON  SAVSY,LSAVSY,LLODBI,10,'SAVESYSTEM'
	.DW  DOLIT,RAMBOT,DUPP,DUPP,SPLIT,SWAP,EMIT,EMIT
	.DW  DOLIT,UPP,SPLIT,SWAP,EMIT,EMIT
	.DW  CP,AT,DUPP,SPLIT,SWAP,EMIT,EMIT
	.DW  USOFF,AT,SPLIT,SWAP,EMIT,EMIT
	.DW  SWAP,SUBB,SAVBI
		;(CP-RAMBOT) bytes starting from RAMBOT
	.DW  DOLIT,UPP,USOFF,AT,SAVBI
		;USOFF bytes starting from UPP
	.DW  EXIT


;LOADSYSTEM ( -- )
;       Reload the precompiled dictionary extension and USER area
;       via SerialCom

	@COLON  LODSY,LLODSY,LSAVSY,10,'LOADSYSTEM'
	.DW  VARAD,AT,PACE                     ;keep in mind & trigger
	.DW  KEY,KEY,JOIN                      ;get RAMBOT
	.DW  DOLIT,RAMBOT,EQUAL,QBRAN,LODSY1   ;test & refuse
	.DW  KEY,KEY,JOIN                      ;get UPP
	.DW  DOLIT,UPP,EQUAL,QBRAN,LODSY1      ;test & refuse
	.DW  PACE                              ;order next bytes
	.DW  KEY,KEY,JOIN                      ;get CP
	.DW  KEY,KEY,JOIN                      ;get USOFF
	.DW  SWAP,DOLIT,RAMBOT,SWAP,OVER       ;get dictionary
	.DW  SUBB,LODBI
	.DW  DOLIT,UPP,SWAP,LODBI              ;get user area
	.DW  VARAD,AT,MIN,VARAD,STORE          ;use biggest VARAD offset
	.DW  EXIT
LODSY1: .DW  DOLIT,27,EMIT,EXIT


; QUIT  ( -- )
;       Reset return stack pointer and start text interpreter.

	@COLON  QUIT,LQUIT,LLODSY,4,'QUIT'
	.DW  RZERO,AT,RPSTO      ;reset return stack pointer
QUIT1:  .DW  LBRAC           ;start interpretation
QUIT2:  .DW  QUERY           ;get input
	.DW  DOLIT,EVAL,CATCH,QDUP   ;evaluate input
	.DW  QBRAN,QUIT2     ;continue till error
	.DW  TPROM,AT,SWAP       ;save input device
	.DW  CONSO,NULLS,OVER,XORR   ;?display error message
	.DW  QBRAN,QUIT3
	.DW  SPACE,COUNT,TYPEE   ;error message
	@DISTR  DOTQP,3,' ? '         ;error prompt
QUIT3:  .DW  DOLIT,DOTOK,XORR    ;?file input
	.DW  QBRAN,QUIT4
	.DW  DOLIT,ERR,EMIT      ;file error, tell host
QUIT4:  .DW  PRESE           ;some cleanup
	.DW  BRAN,QUIT1


;; The compiler

; '     ( -- ca )
;       Search context vocabularies for the next word in input stream.

	@COLON  TICK,LTICK,LQUIT,1,"'"
	.DW  TOKEN,NAMEQ     ;?defined
	.DW  QBRAN,TICK1
	.DW  EXIT            ;yes, push code address
TICK1:  .DW  THROW           ;no, error


; ALLOT ( n -- )
;       Allocate n bytes to the dictionary.

	@COLON  ALLOT,LALLOT,LTICK,5,'ALLOT'
	.DW  CP,PSTOR,EXIT       ;adjust dictionary pointer


; ,     ( w -- )
;       Compile an integer into the dictionary.

	@COLON  COMMA,LCOMMA,LALLOT,1,','
	.DW  HERE,DUPP,CELLP     ;cell boundary
	.DW  CP,STORE,STORE,EXIT ;adjust dictionary pointer, compile


; [COMPILE] ( -- ; <string> )
;       Compile the next immediate word into code dictionary.

	@COLON  BCOMP,LBCOMP,LCOMMA,IMEDD+9,'[COMPILE]'
	.DW  TICK,COMMA,EXIT


; COMPILE ( -- )
;       Compile the next address in colon list to code dictionary.

	@COLON  COMPI,LCOMPI,LBCOMP,COMPO+7,'COMPILE'
	.DW  RFROM,DUPP,AT,COMMA ;compile address
	.DW  CELLP,TOR,EXIT      ;adjust return address


; LITERAL ( w -- )
;       Compile tos to code dictionary as an integer literal.

	@COLON  LITER,LLITER,LCOMPI,IMEDD+7,'LITERAL'
	.DW  COMPI,DOLIT,COMMA,EXIT


; $,"   ( -- )
;       Compile a literal string up to next " .

	@COLON  STRCQ,LSTRCQ,LLITER,3,'$,"'
	.DW  DOLIT,'"',WORDD     ;move string to code dictionary
	.DW  COUNT,PLUS,ALGND    ;calculate aligned end of string
	.DW  CP,STORE,EXIT       ;adjust the dictionary pointer


; RECURSE ( -- )
;       Make the current word available for compilation.
;       (compiles the ca of the latest word
;       into it's own code in the code dictionary)

	@COLON  RECUR,LRECUR,LSTRCQ,IMEDD+7,'RECURSE'
	.DW  LAST,AT,NAMET,COMMA,EXIT


;; Structures

; FOR   ( -- a )
;       Start a FOR-NEXT loop structure in a colon definition.

	@COLON  FOR,LFOR,LRECUR,IMEDD+3,'FOR'
	.DW  COMPI,TOR,HERE,EXIT


; DO    ( -- a )
;       Alias of "FOR"
;
	@COLON  DO,LDO,LFOR,IMEDD+2,'DO'
	.DW  COMPI,TOR,HERE,EXIT


; BEGIN ( -- a )
;       Start an infinite or indefinite loop structure.

	@COLON  BEGIN,LBEGIN,LDO,IMEDD+5,'BEGIN'
	.DW  HERE,EXIT


; NEXT  ( a -- )
;       Terminate a FOR-NEXT loop structure.

	@COLON  NEXT,LNEXT,LBEGIN,IMEDD+4,'NEXT'
	.DW  COMPI,DONXT,COMMA,EXIT


; ZLOOP ( a -- )
;       Alias of NEXT
;       similar to Forth-83 "LOOP", but always counts down to ZERO !

	@COLON  ZLOOP,LZLOOP,LNEXT,IMEDD+5,'ZLOOP'
	.DW  COMPI,DONXT,COMMA,EXIT


; UNTIL ( a -- )
;       Terminate a BEGIN-UNTIL indefinite loop structure.

	@COLON  UNTIL,LUNTIL,LZLOOP,IMEDD+5,'UNTIL'
	.DW  COMPI,QBRAN,COMMA,EXIT


; AGAIN ( a -- )
;       Terminate a BEGIN-AGAIN infinite loop structure.

	@COLON  AGAIN,LAGAIN,LUNTIL,IMEDD+5,'AGAIN'
	.DW  COMPI,BRAN,COMMA,EXIT


; IF    ( -- A )
;       Begin a conditional branch structure.

	@COLON  IFF,LIFF,LAGAIN,IMEDD+2,'IF'
	.DW  COMPI,QBRAN,HERE
	.DW  DOLIT,0,COMMA,EXIT


; AHEAD ( -- A )
;       Compile a forward branch instruction.

	@COLON  AHEAD,LAHEAD,LIFF,IMEDD+5,'AHEAD'
	.DW  COMPI,BRAN,HERE,DOLIT,0,COMMA,EXIT


; REPEAT ( A a -- )
;       Terminate a BEGIN-WHILE-REPEAT indefinite loop.

	@COLON  REPEA,LREPEA,LAHEAD,IMEDD+6,'REPEAT'
	.DW  AGAIN,HERE,SWAP,STORE,EXIT


; THEN  ( A -- )
;       Terminate a conditional branch structure.

	@COLON  THENN,LTHENN,LREPEA,IMEDD+4,'THEN'
	.DW  HERE,SWAP,STORE,EXIT


; AFT   ( a -- a A )
;       Jump to THEN in a FOR-AFT-THEN-NEXT loop the first time through.

	@COLON  AFT,LAFT,LTHENN,IMEDD+3,'AFT'
	.DW  DROP,AHEAD,BEGIN,SWAP,EXIT


; ELSE  ( A -- A )
;       Start the false clause in an IF-ELSE-THEN structure.

	@COLON  ELSEE,LELSEE,LAFT,IMEDD+4,'ELSE'
	.DW  AHEAD,SWAP,THENN,EXIT


; WHILE ( a -- A a )
;       Conditional branch out of a BEGIN-WHILE-REPEAT loop.

	@COLON  WHILE,LWHILE,LELSEE,IMEDD+5,'WHILE'
	.DW  IFF,SWAP,EXIT


; ABORT" ( -- ; <string> )
;        Conditional abort with an error message.

	@COLON  ABRTQ,LABRTQ,LWHILE,IMEDD+6,'ABORT"'
	.DW  COMPI,ABORQ,STRCQ,EXIT


; $"    ( -- ; <string> )
;       Compile an inline string literal.

	@COLON  STRQ,LSTRQ,LABRTQ,IMEDD+2,'$"'
	.DW  COMPI,STRQP,STRCQ,EXIT


; ."    ( -- ; <string> )
;       Compile an inline string literal to be typed out at run time.

	@COLON  DOTQ,LDOTQ,LSTRQ,IMEDD+2,'."'
	.DW  COMPI,DOTQP,STRCQ,EXIT


;; Name compiler

; ?UNIQUE ( a -- a )
;       Display a warning message if the word already exists.

	@COLON  UNIQU,LUNIQU,LDOTQ,7,'?UNIQUE'
	.DW  DUPP,NAMEQ      ;?name exists
	.DW  QBRAN,UNIQ1     ;redefinitions are OK
	@DISTR  DOTQP,7,' reDef '     ;but warn the user
	.DW  OVER,COUNT,TYPEE    ;just in case its not planned
UNIQ1:  .DW  DROP,EXIT


; $,n   ( na -- )
;       Build a new dictionary name using the string at na.

;       legth and name were put into memory from "na" by "TOKEN" and "PACK$"

	@COLON  SNAME,LSNAME,LUNIQU,3,'$,n'
	.DW  DUPP,CAT       ;?null input
	.DW  QBRAN,PNAM1
	.DW  UNIQU          ;?redefinition
	.DW  DUPP,LAST,STORE     ;save na for vocabulary link
	 ;align code address behind the new header
	.DW  DUPP,DUPP,CAT,PLUS,DOLIT,1,PLUS,ALGND,DUPP,CP,STORE
	.DW  SWAP           ;at this point "na" is TOS
						;and the new "ca" second on stack
	.DW  CELLM          ;link address is TOS
	.DW  CRRNT,AT,AT,OVER,STORE ; put the actual link address into "la"
	.DW  CELLM          ;adjust to memory cell for code pointer
	.DW  STORE          ;save code pointer
	.DW  EXIT
PNAM1:  @DISTR  STRQP,5,' name'       ;null input
	.DW  THROW


;; FORTH compiler

; $COMPILE ( a -- )
;       Compile next word to code dictionary as a token or literal.

	@COLON  SCOMP,LSCOMP,LSNAME,8,'$COMPILE'
	.DW  NAMEQ,QDUP      ;?defined
	.DW  QBRAN,SCOM2
	.DW  AT,DOLIT,IMEDD,ANDD ;?immediate
	.DW  QBRAN,SCOM1
	.DW  EXECU,EXIT      ;its immediate, execute
SCOM1:  .DW  COMMA,EXIT      ;its not immediate, compile
SCOM2:  .DW  TNUMB,ATEXE     ;try to convert to number
	.DW  QBRAN,SCOM3
	.DW  LITER,EXIT      ;compile number as integer
SCOM3:  .DW  THROW           ;error


; OVERT ( -- )
;       Link a new word into the current vocabulary.

	@COLON  OVERT,LOVERT,LSCOMP,5,'OVERT'
	.DW  LAST,AT,CRRNT,AT,STORE,EXIT


; ;     ( -- )
;       Terminate a colon definition.

	@COLON  SEMIS,LSEMIS,LOVERT,IMEDD+COMPO+1,';'
	.DW  COMPI,EXIT,LBRAC,OVERT,EXIT


; ]     ( -- )
;       Start compiling the words in the input stream.

	@COLON  RBRAC,LRBRAC,LSEMIS,1,']'
	.DW  DOLIT,SCOMP,TEVAL,STORE,EXIT


; call, ( ca -- )
;       Assemble a call instruction to ca.

	@COLON  CALLC,LCALLC,LRBRAC,5,'call,'
	.DW  DOLIT,$1200,COMMA  ;Direct Threaded Code
				;NOP LJMP in Intel-style byte order
	.DW  COMMA,EXIT


; :     ( -- ; <string> )
;       Start a new colon definition using next word as its name.

	@COLON  COLON,LCOLON,LCALLC,1,':'
	.DW  TOKEN,SNAME,DOLIT,DOLST,FLIP
	.DW  CALLC,RBRAC,EXIT


; IMMEDIATE ( -- )
;       Make the last compiled word an immediate word.

	@COLON  IMMED,LIMMED,LCOLON,9,'IMMEDIATE'
	.DW  DOLIT,IMEDD,LAST,AT,AT,ORR
	.DW  LAST,AT,STORE,EXIT


;; Defining words

; CREATE ( -- ; <string> )
;       Compile a new array entry without allocating code space.

	@COLON  CREAT,LCREAT,LIMMED,6,'CREATE'
	.DW  TOKEN,SNAME,OVERT
	.DW  DOLIT,DOLST,FLIP,CALLC
	.DW  COMPI,DOVAR,EXIT


; CONSTANT ( w -- ; <string> )
;       Compile an new constant with value w

	@COLON  CONST,LCONST,LCREAT,8,'CONSTANT'
	.DW  TOKEN,SNAME,OVERT  ;compile header
	.DW  DOLIT,DOLST,FLIP,CALLC
	.DW  COMPI,DOLIT,COMMA  ;compile runtime code
	.DW  COMPI,EXIT,EXIT


; USER  ( u -- ; <string> )
;       Compile a new user variable.

	@COLON  USER,LUSER,LCONST,4,'USER'
	.DW  TOKEN,SNAME,OVERT
	.DW  DOLIT,DOLST,FLIP,CALLC
	.DW  COMPI,DOUSE,COMMA,EXIT


; USA  ( -- ; <string> )
;       Compile a new user variable with automatic offset increment.

	@COLON  USA,LUSA,LUSER,3,'USA'

	.DW  USOFF,DUPP,AT,SWAP     ;( -- w a)
	.DW  OVER,CELLP,SWAP,STORE  ;w+2 --> USOFF ; ( -- w)
	.DW  USER,EXIT


; VARIABLE ( -- ; <string> )
;       Compile a new variable initialized to 0.

	@COLON  VARIA,LVARIA,LUSA,8,'VARIABLE'
;  .DW  CREAT,DOLIT,0,COMMA,EXIT (original version used inline RAM)
	.DW  VARAD,DUPP,AT,SWAP
	.DW  OVER,CELLM,SWAP,STORE      ;VARAD-2 --> VARAD ; ( -- w)
	.DW  DUPP,DOLIT,0,SWAP,STORE    ;initialize Variable = 0
	.DW  TOKEN,SNAME,OVERT          ;compile header
	.DW  DOLIT,DOLST,FLIP,CALLC
	.DW  COMPI,DOLIT,COMMA          ;compile runtime code
	.DW  COMPI,EXIT,EXIT


;FUNCTION ( a -- ; <string> )
;       Compile a call for external function.
;       The executable machine code must begin at address a.
;       Back to eForth at the first "ret" instruction found.
;       Free use of 8031 Accu, Register Banks 2 and 3,
;       hardware stack, if left with aligned stackpointer.
;       Free use of SFR's, memory, subroutines and interrupts
;       as long as not destroing eForth system & setup.
;       Internal RAM lower than $10 is forbidden, because eForth
;       pointers are stored there.
;       Hardware stack is in internal RAM, from $40 growing upwards.
;       This area is forbidden, too.
;       When calling "FUNCTION", automatically register bank #2
;       is activated; after "ret" selection is done back to bank #1.

	@COLON  FUNCT,LFUNCT,LVARIA,8,'FUNCTION'
	.DW  TOKEN,SNAME,OVERT         ;compile primary header
	.DW  DOLIT,$C2D3,FLIP,COMMA    ;compile "clr bit D3" (PSW,RS0)
	.DW  DOLIT,$D2D4,FLIP,COMMA    ;"set bit D4" (PSW,RS1)
				       ;(Select register bank #2)
	.DW  DOLIT,$0012,FLIP,COMMA    ;"nop,lcall"
	.DW  FLIP,COMMA                ;compile a
	.DW  DOLIT,$D2D3,FLIP,COMMA    ;"set bit D3" (PSW,RS0)
	.DW  DOLIT,$C2D4,FLIP,COMMA    ;"clr bit D4" (PSW,RS1)
				       ;(Re-Select register bank #1)
	.DW  DOLIT,$0002,FLIP,COMMA    ; "nop,ljmp"
	.DW  DOLIT,RETURN,FLIP,COMMA
	.DW  EXIT


;; Tools

; _TYPE ( b u -- )
;       Display a string. Filter non-printing characters.

	@COLON  UTYPE,LUTYPE,LFUNCT,5,'_TYPE'
	.DW  TOR         ;start count down loop
	.DW  BRAN,UTYP2      ;skip first pass
UTYP1:  .DW  DUPP,CAT,TCHAR,EMIT ;display only printable
	.DW  DOLIT,1,PLUS        ;increment address
UTYP2:  .DW  DONXT,UTYP1     ;loop till done
	.DW  DROP,EXIT


; dm+   ( a u -- a )
;       Dump u bytes from , leaving a+u on the stack.

	@COLON  DMP,LDMP,LUTYPE,3,'dm+'
	.DW  OVER,DOLIT,4,UDOTR  ;display address
	.DW  SPACE,TOR       ;start count down loop
	.DW  BRAN,PDUM2      ;skip first pass
PDUM1:  .DW  DUPP,CAT,DOLIT,3,UDOTR  ;display numeric data
	.DW  DOLIT,1,PLUS        ;increment address
PDUM2:  .DW  DONXT,PDUM1     ;loop till done
	.DW  EXIT


; DUMP  ( a u -- )
;       Dump u bytes from a, in a formatted manner.

	@COLON  DUMP,LDUMP,LDMP,4,'DUMP'
	.DW  BASE,AT,TOR,HEX     ;save radix, set hex
	.DW  DOLIT,16,SLASH      ;change count to lines
	.DW  TOR         ;start count down loop
DUMP1:  .DW  CR,DOLIT,16,DDUP,DMP    ;display numeric
	.DW  ROT,ROT
	.DW  SPACE,SPACE,UTYPE   ;display printable characters
	.DW  NUFQ,INVER      ;user control
	.DW  QBRAN,DUMP2
	.DW  DONXT,DUMP1     ;loop till done
	.DW  BRAN,DUMP3
DUMP2:  .DW  RFROM,DROP      ;cleanup loop stack, early exit
DUMP3:  .DW  DROP,RFROM,BASE,STORE   ;restore radix
	.DW  EXIT


; .S    ( ... -- ... )
;       Display the contents of the data stack.

	@COLON  DOTS,LDOTS,LDUMP,2,'.S'
	.DW  CR,DEPTH        ;stack depth
	.DW  TOR         ;start count down loop
	.DW  BRAN,DOTS2      ;skip first pass
DOTS1:  .DW  RAT,PICK,DOT        ;index stack, display contents
DOTS2:  .DW  DONXT,DOTS1     ;loop till done
	@DISTR  DOTQP,4,' <sp'
	.DW  EXIT


; !CSP  ( -- )
;       Save stack pointer in CSP for error checking.

	@COLON  STCSP,LSTCSP,LDOTS,4,'!CSP'
	.DW  SPAT,CSP,STORE,EXIT ;save pointer


; ?CSP  ( -- )
;       Abort if stack pointer differs from that saved in CSP.

	@COLON  QCSP,LQCSP,LSTCSP,4,'?CSP'
	.DW  SPAT,CSP,AT,XORR    ;compare pointers
	@DISTR  ABORQ,6,'stacks'      ;abort if different
	.DW  EXIT


;>NAME  ( ca -- na | F )
;       Convert code address to a name address.

	@COLON  TNAME,LTNAME,LQCSP,5,'>NAME'
	.DW  CRRNT           ;vocabulary link
TNAM1:  .DW  CELLP,AT,QDUP       ;check all vocabularies
	.DW  QBRAN,TNAM4
	.DW  DDUP
TNAM2:  .DW  AT,DUPP         ;?last word in a vocabulary
	.DW  QBRAN,TNAM3
	.DW  DDUP,NAMET,XORR     ;compare
	.DW  QBRAN,TNAM3
	.DW  CELLM           ;continue with next word
	.DW  BRAN,TNAM2
TNAM3:  .DW  SWAP,DROP,QDUP
	.DW  QBRAN,TNAM1
	.DW  SWAP,DROP,SWAP,DROP,EXIT
TNAM4:  .DW  DROP,DOLIT,0,EXIT   ;false flag


; .ID   ( na -- )
;       Display the name at address.

	@COLON  DOTID,LDOTID,LTNAME,3,'.ID'
	.DW  QDUP                    ;if zero no name
	.DW  QBRAN,DOTI1
	.DW  COUNT,DOLIT,01FH,ANDD   ;mask lexicon bits
	.DW  UTYPE,EXIT              ;display name string
DOTI1:  @DISTR  DOTQP,9,' {noName}'
	.DW  EXIT


; SEE   ( -- ; <string> )
;       A simple decompiler.

	@COLON  SEE,LSEE,LDOTID,3,'SEE'
	.DW  TICK                ;starting address
	.DW  CR,CELLP
SEE1:   .DW  CELLP,DUPP,AT,DUPP  ;?does it contain a zero
	.DW  QBRAN,SEE2
	.DW  TNAME           ;?is it a name
SEE2:   .DW  QDUP            ;name address or zero
	.DW  QBRAN,SEE3
	.DW  SPACE,DOTID     ;display name
	.DW  BRAN,SEE4
SEE3:   .DW  DUPP,AT,UDOT    ;display number
SEE4:   .DW  NUFQ            ;user control
	.DW  QBRAN,SEE1
	.DW  DROP,EXIT


; WORDS ( -- )
;       Display the names in the context vocabulary.

	@COLON  WORDS,LWORDS,LSEE,5,'WORDS'
	.DW  CR,CNTXT,AT     ;only in context
WORS1:  .DW  AT,QDUP         ;?at end of list
	.DW  QBRAN,WORS2
	.DW  DUPP,SPACE,DOTID    ;display a name
	.DW  CELLM,NUFQ      ;user control
	.DW  QBRAN,WORS1
	.DW  DROP
WORS2:  .DW  EXIT


; FORGET   ( --; <string> )
;       Erase the word with name <string> and all following from dictionary
;       Reset current dictionary pointer to last remaining word.
;       More sophisticated dictionary handling not yet implemented.

	@COLON  FORGE,LFORGE,LWORDS,6,'FORGET'
	.DW  TICK                     ;return ca or throw
	.DW  TNAME                    ;return na or FALSE
	.DW  CELLM,CELLM,DUPP,CP,STORE;set CP to begin of erased word
	.DW  CELLP,AT,DUPP,LAST,STORE ;set LAST to na of word before in dict.
	.DW  OVERT,EXIT


;; Hardware reset

; VER   ( -- n )
;       Return the version number of this implementation.

	@COLON  VERSN,LVERSN,LFORGE,3,'VER'
	.DW  DOLIT,VER*256+EXT,EXIT


; hi    ( -- )
;       Display the sign-on message of eForth.

	@COLON  HI,LHI,LVERSN,2,'hi'
	.DW  STOIO,CR        ;initialize I/O
	@DISTR  DOTQP,8,'eForth v'    ;model
	.DW  BASE,AT,HEX     ;save radix
	.DW  VERSN,BDIGS,DIG,DIG
	.DW  DOLIT,'.',HOLD
	.DW  DIGS,EDIGS,TYPEE    ;format version number
	.DW  BASE,STORE,CR,EXIT  ;restore radix



; Table of values to be written as initial USER variables

IZERO:  .DW  SPP        ;SP0
	.DW  RPP        ;RP0
	.DW  QRX        ;'?KEY
	.DW  TXSTO      ;'EMIT
	.DW  ACCEP      ;'EXPECT
	.DW  KTAP       ;'TAP
	.DW  TXSTO      ;'ECHO
	.DW  DOTOK      ;'PROMPT
	.DW  BASEE      ;BASE
	.DW  0          ;tmp
	.DW  0          ;SPAN
	.DW  0          ;>IN
	.DW  0          ;#TIB
	.DW  TIBB       ;TIB
	.DW  0          ;CSP
	.DW  INTER      ;'EVAL
	.DW  NUMBQ      ;'NUMBER
	.DW  0          ;HLD
	.DW  0          ;HANDLER
	.DW  0          ;VARAD
	.DW  0          ;USOFF
	.DW  0          ;CONTEXT pointer
	.BLOCK 16,0     ;vocabulary stack
	.DW  0          ;CURRENT pointer
	.DW  0          ;vocabulary link pointer
	.DW  RAMBOT     ;CP == conventional dictionary pointer
	.DW  RAMBOT + 4 ;NP == pointer to the actual name field
ILAST:                  ;the following items have to be calculated
			;separately, because LASTN=#COLD+4 isn't yet known

; 'BOOT ( -- a )
;       The application startup vector.

	@COLON  TBOOT,LTBOOT,LHI,5,"'BOOT"
	.DW  DOVAR
	.DW  HI          ;application to boot


; COLDINIT ( -- )
;       initialisation of USER area

	@CODE CINIT,LCINIT,LTBOOT,8,'COLDINIT'
	mov DPTR,#UPP
	mov UPL,DPL
	mov UPH,DPH                     ;initialize USER area in RAM
	mov REG0,#0                     ;Offset from IZERO
	mov REG1,#(ILAST-IZERO)         ;count variable
CINIT1: mov A,REG0                      ;move memory block
	mov DPTR,#IZERO
	movc A,@A+DPTR
	mov DPL,UPL
	mov DPH,UPH
	movx @DPTR,A
	inc REG0
	inc UPL
	djnz REG1,CINIT1

	mov DPTR,#UPP
	mov UPL,DPL                     ;reset to inital value

	lcall GETLCOLD                  ;recapitulate last dictionary entry
					;LCOLD = last na
	mov REG0,DPL
	mov REG1,DPH
	mov DPTR,#ULAST                 ; na to USER variable LAST
	mov A,REG0
	movx @DPTR,A
	inc DPTR
	mov A,REG1
	movx @DPTR,A
	mov DPTR,#UVFRTH                ;na to USER variable VFRTH
	mov A,REG0
	movx @DPTR,A
	inc DPTR
	mov A,REG1
	movx @DPTR,A

	mov DPTR,#UVL                   ;zero initialisation to voc link
	clr A
	movx @DPTR,A
	inc DPTR
	movx @DPTR,A

	mov DPTR,#VARAM         ;initialisation of "VARIABLE"s pfa RAM
	mov REG0,DPL
	mov REG1,DPH
	mov DPTR,#UVARAD
	mov A,REG0
	movx @DPTR,A
	inc DPTR
	mov A,REG1
	movx @DPTR,A

	mov DPTR,#(UVL-UPP+2)   ;offset of next free USER cell
	mov REG0,DPL
	mov REG1,DPH
	mov DPTR,#UUSOFF
	mov A,REG0
	movx @DPTR,A
	inc DPTR
	mov A,REG1
	movx @DPTR,A

	mov DPTR,#SFRCODE       ;initialize RAM with recompilable code loop
				;to be used by SFR!
	mov A,#$F5
	movx @DPTR,A            ;opcode "mov direct,A"
	inc DPTR
	mov A,#$FF              ;dummy, to be dynamically recompiled
	movx @DPTR,A            ;at runtime
	inc DPTR
	mov A,#$22              ;opcode "ret"
	movx @DPTR,A
				;the following to be used by SFR@
	inc DPTR
	mov A,#$E5              ;opcode "mov A,direct"
	movx @DPTR,A
	inc DPTR
	mov A,#$FF
	movx @DPTR,A
	inc DPTR
	mov A,#$22
	movx @DPTR,A

	ljmp RETURN


; COLD  ( -- )
;       The hilevel cold start sequence.

	@COLON  COLD,LCOLD,LCINIT,4,'COLD'
COLD1:  .DW  CINIT
;        .DW  PRESE           ;initialize stack and TIB
	.DW  TBOOT,ATEXE     ;application boot
	.DW  FORTH,CNTXT,AT,DUPP ;initialize search order
	.DW  CRRNT,DSTOR,OVERT
	.DW  QUIT            ;start interpretation
	.DW  BRAN,COLD1      ;just in case


GETLCOLD:
	mov DPTR,#LCOLD     ;tricking out the forward references
	ret
GETCOLD1:
	mov DPTR,#COLD1
	ret

SENDA:


	jnb SCON.1,TXOUT1
	clr SCON.1  ;Flag lschen
TXOUT1: mov SBUF,A
TXOUT2:
;Sendeschleife
; warten bis TX-Interrupt-Flag gesetzt
	jnb SCON.1,TXOUT2
	clr SCON.1  ;Flag wieder lschen
	ret


	.ORG $1FFF      ;filling the rest of the EPROM with $FF
	.DB $FF
	.ENDS

	.END

;==============================================================
